# inputs file

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt 

def set_inputs():

    #Step 1: User Inputs for scenario
    resource_year = 2013
    atb_years = [
                2022,
                # 2025,
                # 2030,
                # 2035
                ]
    policy = {
        'option 1': {'Wind ITC': 0, 'Wind PTC': 0, "H2 PTC": 0},
        # 'option 2': {'Wind ITC': 26, 'Wind PTC': 0, "H2 PTC": 0},
        # 'option 3': {'Wind ITC': 0, 'Wind PTC': 0.003, "H2 PTC": 0},
        # 'option 4': {'Wind ITC': 0, 'Wind PTC': 0.026, "H2 PTC": 0},
        # 'option 5': {'Wind ITC': 0, 'Wind PTC': 0.003, "H2 PTC": 0.6},
        'option 6': {'Wind ITC': 0, 'Wind PTC': 0.026, "H2 PTC": 3},
    }

    sample_site['year'] = resource_year
    useful_life = 30
    critical_load_factor = 1
    run_reopt_flag = False
    custom_powercurve = True
    storage_used = True
    battery_can_grid_charge = True
    grid_connected_hopp = False
    interconnection_size_mw = 1000
    electrolyzer_size = 1000

def establish_save_output_dict():
    """
    Establishes and returns a 'save_outputs_dict' dict
    for saving the relevant analysis variables for each site.
    """

    save_outputs_dict = dict()
    save_outputs_dict['Site Name'] = list()
    save_outputs_dict['Substructure Technology'] = list()
    save_outputs_dict['ATB Year'] = list()
    save_outputs_dict['Policy Option'] = list()
    save_outputs_dict['Resource Year'] = list()
    save_outputs_dict['Turbine Model'] = list()
    save_outputs_dict['Critical Load Factor'] = list()
    save_outputs_dict['System Load (kW)'] = list()
    save_outputs_dict['Useful Life'] = list()
    save_outputs_dict['Wind PTC ($/kW)'] = list()
    save_outputs_dict['H2 PTC ($/kg)'] = list()
    save_outputs_dict['ITC (%)'] = list()
    save_outputs_dict['Discount Rate (%)'] = list()
    save_outputs_dict['Debt Equity'] = list()
    save_outputs_dict['Hub Height (m)'] = list()
    save_outputs_dict['Storage Enabled'] = list()
    save_outputs_dict['Wind Cost ($/kW)'] = list()
    save_outputs_dict['Solar Cost ($/kW)'] = list()
    save_outputs_dict['Storage Cost ($/kW)'] = list()
    save_outputs_dict['Storage Cost ($/kWh)'] = list()
    save_outputs_dict['Electrolyzer Cost ($/kW)'] = list()
    save_outputs_dict['Storage Hours'] = list()
    save_outputs_dict['Wind built (MW)'] = list()
    save_outputs_dict['Solar built (MW)'] = list()
    save_outputs_dict['Storage built (MW)'] = list()
    save_outputs_dict['Storage built (MWh)'] = list()
    save_outputs_dict['Electrolyzer built (MW)'] = list()
    save_outputs_dict['Battery Can Grid Charge'] = list()
    save_outputs_dict['Grid Connected HOPP'] = list()
    save_outputs_dict['Built Interconnection Size (MW)'] = list()
    save_outputs_dict['Wind Total Installed Cost ($)'] = list()
    save_outputs_dict['Desal Total Installed Cost ($)'] = list()
    save_outputs_dict['HVDC Total Installed Cost ($)'] = list()
    save_outputs_dict['Pipeline Total Installed Cost ($)'] = list()
    save_outputs_dict['Electrolyzer Total Installed Capital Cost ($)'] = list()
    save_outputs_dict['Total Plant Capital Cost HVDC ($)'] = list()
    save_outputs_dict['Total Plant Capital Cost Pipeline ($)'] = list()
    save_outputs_dict['Wind Total Operating Cost ($)'] = list()
    save_outputs_dict['Wind Fixed Operating Cost ($)'] = list()
    save_outputs_dict['Electrolyzer Total Operating Cost ($)'] = list()
    save_outputs_dict['Desal Total Operating Cost ($)'] = list()
    save_outputs_dict['HVDC Total Operating Cost ($)'] = list()
    save_outputs_dict['Pipeline Total Operating Cost ($)'] = list()
    save_outputs_dict['Total Plant Operating Cost HVDC ($)'] = list()
    save_outputs_dict['Total Plant Operating Cost Pipeline ($)'] = list()
    save_outputs_dict['Real LCOE ($/MWh)'] = list()
    save_outputs_dict['Nominal LCOE ($/MWh)'] = list()
    save_outputs_dict['Total Annual H2 production (kg)'] = list()
    save_outputs_dict['H2 total tax credit ($)'] = list()
    save_outputs_dict['Total ITC HVDC scenario ($)'] = list()
    save_outputs_dict['Total ITC Pipeline scenario ($)'] = list()
    save_outputs_dict['TLCC Wind ($)'] = list()
    save_outputs_dict['TLCC H2 ($)'] = list()
    save_outputs_dict['TLCC Desal ($)'] = list()
    save_outputs_dict['TLCC HVDC ($)'] = list()
    save_outputs_dict['TLCC Pipeline ($)'] = list()
    save_outputs_dict['LCOH Wind contribution ($/kg)'] = list()
    save_outputs_dict['LCOH H2 contribution ($/kg)'] = list()
    save_outputs_dict['LCOH Desal contribution ($/kg)'] = list()
    save_outputs_dict['LCOH HVDC contribution ($/kg)'] = list()
    save_outputs_dict['LCOH Pipeline contribution ($/kg)'] = list()    
    save_outputs_dict['H_LCOE HVDC scenario no operating costs (uses LCOE calculator) ($/kg)'] = list()
    save_outputs_dict['H_LCOE HVDC scenario w/ operating costs (uses LCOE calculator) ($/kg)'] = list()
    save_outputs_dict['H_LCOE Pipeline scenario no operating costs (uses LCOE calculator) ($/kg)'] = list()
    save_outputs_dict['H_LCOE Pipeline scenario w/ operating costs (uses LCOE calculator) ($/kg)'] = list()
    save_outputs_dict['Levelized Cost/kg H2 HVDC scenario (CF Method - using annual cashflows per technology) ($/kg)'] = list()
    save_outputs_dict['Levelized Cost/kg H2 Pipeline scenario (CF Method - using annual cashflows per technology) ($/kg)'] = list()
    save_outputs_dict['Gut-Check Cost/kg H2 Pipeline scenario (non-levelized, includes elec if used) ($/kg)'] = list()
    save_outputs_dict['Gut-Check Cost/kg H2 HVDC scenario (non-levelized, includes elec if used) ($/kg)'] = list()
    save_outputs_dict['Grid Connected HOPP'] = list()
    save_outputs_dict['HOPP Total Electrical Generation (kWh)'] = list()
    save_outputs_dict['Total Yearly Electrical Generation used by Electrolyzer (kWh)'] = list()
    save_outputs_dict['Wind Capacity Factor (%)'] = list()
    save_outputs_dict['Electrolyzer Capacity Factor (%)'] = list()
    save_outputs_dict['HOPP Annual Energy Shortfall (kWh)'] = list()
    save_outputs_dict['HOPP Annual Energy Curtailment (kWh)'] = list()
    save_outputs_dict['Battery Generation (kWh)'] = list()
    save_outputs_dict['Electricity to Grid (kWh)'] = list()
    
    return save_outputs_dict

def save_the_things():
    save_outputs_dict = dict()
    save_outputs_dict['Site Name'] = (site_name)
    save_outputs_dict['Substructure Technology'] = (site_df['Substructure technology'])
    save_outputs_dict['ATB Year'] = (atb_year)
    save_outputs_dict['Policy Option'] = (policy_option)
    save_outputs_dict['Resource Year'] = (resource_year)
    save_outputs_dict['Turbine Model'] = (turbine_model)
    save_outputs_dict['Critical Load Factor'] = (critical_load_factor)
    save_outputs_dict['System Load (kW)'] = (kw_continuous)
    save_outputs_dict['Useful Life'] = (useful_life)
    save_outputs_dict['Wind PTC ($/kW)'] = (scenario['Wind PTC'])
    save_outputs_dict['H2 PTC ($/kg)'] = (scenario['H2 PTC'])
    save_outputs_dict['ITC (%)'] = (scenario['Wind ITC'])
    save_outputs_dict['Discount Rate (%)'] = (discount_rate*100)
    save_outputs_dict['Debt Equity'] = (debt_equity_split)
    save_outputs_dict['Hub Height (m)'] = (tower_height)
    save_outputs_dict['Storage Enabled'] = (storage_used)
    save_outputs_dict['Wind Cost ($/kW)'] = (wind_cost_kw)
    save_outputs_dict['Solar Cost ($/kW)'] = (solar_cost_kw)
    save_outputs_dict['Storage Cost ($/kW)'] = (storage_cost_kw)
    save_outputs_dict['Storage Cost ($/kWh)'] = (storage_cost_kwh)
    save_outputs_dict['Electrolyzer Cost ($/kW)'] = (electrolyzer_capex_kw)
    save_outputs_dict['Storage Hours'] = (storage_hours)
    save_outputs_dict['Wind built (MW)'] = (wind_size_mw)
    save_outputs_dict['Solar built (MW)'] = (solar_size_mw)
    save_outputs_dict['Storage built (MW)'] = (storage_size_mw)
    save_outputs_dict['Storage built (MWh)'] = (storage_size_mwh)
    save_outputs_dict['Electrolyzer built (MW)'] = (electrolyzer_size_mw)
    save_outputs_dict['Battery Can Grid Charge'] = (battery_can_grid_charge)
    save_outputs_dict['Grid Connected HOPP'] = (grid_connected_hopp)
    save_outputs_dict['Built Interconnection Size (MW)'] = (hybrid_plant.interconnect_kw)
    save_outputs_dict['Wind Total Installed Cost ($)'] = (total_hopp_installed_cost)
    save_outputs_dict['Desal Total Installed Cost ($)'] = (total_desal_cost)
    save_outputs_dict['HVDC Total Installed Cost ($)'] = (total_export_system_cost)
    save_outputs_dict['Pipeline Total Installed Cost ($)'] = (total_h2export_system_cost)
    save_outputs_dict['Electrolyzer Total Installed Capital Cost ($)'] = (electrolyzer_total_capital_cost)
    save_outputs_dict['Total Plant Capital Cost HVDC ($)'] = (total_system_installed_cost_hvdc)
    save_outputs_dict['Total Plant Capital Cost Pipeline ($)'] = (total_system_installed_cost_pipeline)
    save_outputs_dict['Wind Total Operating Cost ($)'] = (annual_operating_cost_wind)
    save_outputs_dict['Wind Fixed Operating Cost ($)'] = (fixed_om_cost_wind)
    save_outputs_dict['Electrolyzer Total Operating Cost ($)'] = (annual_operating_cost_h2)
    save_outputs_dict['Desal Total Operating Cost ($)'] = (annual_operating_cost_desal)
    save_outputs_dict['HVDC Total Operating Cost ($)'] = (total_export_om_cost)
    save_outputs_dict['Pipeline Total Operating Cost ($)'] = (opex_pipeline)
    save_outputs_dict['Total Plant Operating Cost HVDC ($)'] = (total_annual_operating_costs_hvdc)
    save_outputs_dict['Total Plant Operating Cost Pipeline ($)'] = (total_annual_operating_costs_pipeline)
    save_outputs_dict['Real LCOE ($/MWh)'] = (lcoe*10)
    save_outputs_dict['Nominal LCOE ($/MWh)'] = (lcoe_nom*10)
    save_outputs_dict['Total Annual H2 production (kg)'] = (H2_Results['hydrogen_annual_output'])
    save_outputs_dict['H2 total tax credit ($)'] = (np.sum(h2_tax_credit))
    save_outputs_dict['Total ITC HVDC scenario ($)'] = (total_itc_hvdc)
    save_outputs_dict['Total ITC Pipeline scenario ($)'] = (total_itc_pipeline)
    save_outputs_dict['TLCC Wind ($)'] = (tlcc_wind_costs)
    save_outputs_dict['TLCC H2 ($)'] = (tlcc_h2_costs)
    save_outputs_dict['TLCC Desal ($)'] = (tlcc_desal_costs)
    save_outputs_dict['TLCC HVDC ($)'] = (tlcc_hvdc_costs)
    save_outputs_dict['TLCC Pipeline ($)'] = (tlcc_pipeline_costs)
    save_outputs_dict['LCOH Wind contribution ($/kg)'] = (LCOH_cf_method_wind)
    save_outputs_dict['LCOH H2 contribution ($/kg)'] = (LCOH_cf_method_h2_costs)
    save_outputs_dict['LCOH Desal contribution ($/kg)'] = (LCOH_cf_method_desal_costs)
    save_outputs_dict['LCOH HVDC contribution ($/kg)'] = (LCOH_cf_method_pipeline)
    save_outputs_dict['LCOH Pipeline contribution ($/kg)'] =  (LCOH_cf_method_hvdc) 
    save_outputs_dict['H_LCOE HVDC scenario no operating costs (uses LCOE calculator) ($/kg)'] = (h_lcoe_no_op_cost_hvdc)
    save_outputs_dict['H_LCOE HVDC scenario w/ operating costs (uses LCOE calculator) ($/kg)'] = (h_lcoe_hvdc)
    save_outputs_dict['H_LCOE Pipeline scenario no operating costs (uses LCOE calculator) ($/kg)'] = (h_lcoe_no_op_cost_pipeline)
    save_outputs_dict['H_LCOE Pipeline scenario w/ operating costs (uses LCOE calculator) ($/kg)'] = (h_lcoe_pipeline)
    save_outputs_dict['Levelized Cost/kg H2 HVDC scenario (CF Method - using annual cashflows per technology) ($/kg)'] = (LCOH_cf_method_total_hvdc)
    save_outputs_dict['Levelized Cost/kg H2 Pipeline scenario (CF Method - using annual cashflows per technology) ($/kg)'] = (LCOH_cf_method_total_pipeline)
    save_outputs_dict['Gut-Check Cost/kg H2 Pipeline scenario (non-levelized, includes elec if used) ($/kg)'] = (gut_check_h2_cost_kg_pipeline)
    save_outputs_dict['Gut-Check Cost/kg H2 HVDC scenario (non-levelized, includes elec if used) ($/kg)'] = (gut_check_h2_cost_kg_hvdc)
    save_outputs_dict['Grid Connected HOPP'] = (grid_connected_hopp)
    save_outputs_dict['HOPP Total Electrical Generation (kWh)'] = (np.sum(hybrid_plant.grid.generation_profile[0:8760]))
    save_outputs_dict['Total Yearly Electrical Generation used by Electrolyzer (kWh)'] = (total_elec_production)
    save_outputs_dict['Wind Capacity Factor (%)'] = (hybrid_plant.wind.capacity_factor)
    save_outputs_dict['Electrolyzer Capacity Factor (%)'] = (H2_Results['cap_factor']*100)
    save_outputs_dict['HOPP Annual Energy Shortfall (kWh)'] = (np.sum(energy_shortfall_hopp))
    save_outputs_dict['HOPP Annual Energy Curtailment (kWh)'] = (np.sum(combined_hybrid_curtailment_hopp))
    save_outputs_dict['Battery Generation (kWh)'] = (np.sum(battery_used))
    save_outputs_dict['Electricity to Grid (kWh)'] = (np.sum(excess_energy))

    return save_outputs_dict


import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit

def _electrolyzer_footprint_data():
    """
    References:
    [1] Bolhui, 2017 https://www.irena.org/-/media/Files/IRENA/Agency/Publication/2020/Dec/IRENA_Green_hydrogen_cost_2020.pdf
        - appears to include BOS
    [2] Bourne, 2017
        - 
    [3] McPHy, 2018 (https://mcphy.com/en/equipment-services/electrolyzers/)
    [4] Air Liquide 2021, Becancour Quebec
    """

    rating_mw = np.array([300, 100, 100, 20]) # [1], [2], [3], [4]
    footprint_sqft = np.array([161500, 37700, 48500, 465000]) # [1], [2], [3], [4]
    sqft_to_m2 = 0.092903
    footprint_m2 = footprint_sqft*sqft_to_m2

    return rating_mw, footprint_m2

def footprint(rating_mw):

    """
    Estimate the area required for the electrolyzer equipment using a linear scaling
    """
    
    footprint_m2 = rating_mw*48000*(1/1E3) # from Singlitico 2021, Table 1 (ratio is given in m2/GW, so a conversion is used here for MW)
    
    return footprint_m2

def _electrolyzer_mass_data():
    """
    References:
    [1] https://www.h-tec.com/en/products/detail/h-tec-pem-electrolyser-me450/me450/
    [2] https://www.nrel.gov/docs/fy19osti/70380.pdf
    """

    rating_mw = np.array([1, 1.25, 0.25, 45E-3, 40E-3, 28E-3, 14E-3, 14.4E-3, 7.2E-7])
    mass_kg = np.array([36E3, 17E3, 260, 900, 908, 858, 682, 275, 250])

    return rating_mw, mass_kg

def _electrolyzer_mass_fit(x, m, b):

    y = m*x + b

    return y

def mass(rating_mw):
    """
    Estimate the electorlyzer mass given the electrolyzer rating based on data.

    Note: the largest electrolyzer data available was for 1.25 MW. Also, given the current fit, the mass goes negative for very small electrolysis systems
    """

    rating_mw_fit, mass_kg_fit = _electrolyzer_mass_data()

    (m, b), pcov = curve_fit(_electrolyzer_mass_fit, rating_mw_fit, mass_kg_fit)

    mass_kg = _electrolyzer_mass_fit(rating_mw, m, b)

    return mass_kg

if __name__ == "__main__":

    fig,ax = plt.subplots(1,2)
    rating_mw, footprint_m2 = _electrolyzer_footprint_data()
    ax[0].scatter(rating_mw, footprint_m2, label="Data points")

    ratings = np.arange(0,1000)
    footprints = footprint(ratings)
    
    ax[0].plot(ratings, footprints, label="Scaling Factor")
    ax[0].set(xlabel="Electrolyzer Rating (MW)", ylabel="Footprint (m$^2$)")
    ax[0].legend(frameon=False)
    print(rating_mw, footprint_m2)

    rating_mw, mass_kg = _electrolyzer_mass_data()
    ax[1].scatter(rating_mw, np.multiply(mass_kg, 1E-3), label="Data points")

    ax[1].plot(ratings, mass(ratings)*1E-3, label="Linear Fit")
    ax[1].set(xlabel="Electrolyzer Rating (MW)", ylabel="Mass (tonnes)")
    ax[1].legend(frameon=False)
    plt.tight_layout()
    plt.show()
    print(rating_mw, np.divide(mass_kg, rating_mw))


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import ticker
import os
import warnings
from greenheart.tools.eco.utilities import ceildiv

# import hopp.tools.hopp_tools as hopp_tools

from greenheart.simulation.technologies.hydrogen.desal.desal_model_eco import (
    RO_desal_eco as RO_desal,
)
from greenheart.simulation.technologies.hydrogen.electrolysis.pem_mass_and_footprint import (
    mass as run_electrolyzer_mass,
)
from greenheart.simulation.technologies.hydrogen.electrolysis.pem_mass_and_footprint import (
    footprint as run_electrolyzer_footprint,
)
from greenheart.simulation.technologies.hydrogen.electrolysis.H2_cost_model import (
    basic_H2_cost_model,
)
from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_costs_Singlitico_model import (
    PEMCostsSingliticoModel,
)

# from hopp.simulation.technologies.hydrogen.electrolysis.run_h2_PEM_eco import run_h2_PEM
from greenheart.simulation.technologies.hydrogen.electrolysis.run_h2_PEM import (
    run_h2_PEM,
)

from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_BOP.PEM_BOP import pem_bop


def run_electrolyzer_physics(
    hopp_results,
    greenheart_config,
    wind_resource,
    design_scenario,
    show_plots=False,
    save_plots=False,
    output_dir="./output/",
    verbose=False,
):

    electrolyzer_size_mw = greenheart_config["electrolyzer"]["rating"]
    electrolyzer_capex_kw = greenheart_config["electrolyzer"]["electrolyzer_capex"]

    # IF GRID CONNECTED
    if greenheart_config["project_parameters"]["grid_connection"]:
        # NOTE: if grid-connected, it assumes that hydrogen demand is input and there is not
        # multi-cluster control strategies. This capability exists at the cluster level, not at the
        # system level.
        if greenheart_config["electrolyzer"]["sizing"]["hydrogen_dmd"] is not None:
            grid_connection_scenario = "grid-only"
            hydrogen_production_capacity_required_kgphr = greenheart_config[
                "electrolyzer"
            ]["sizing"]["hydrogen_dmd"]
            energy_to_electrolyzer_kw = []
        else:
            grid_connection_scenario = "off-grid"
            hydrogen_production_capacity_required_kgphr = []
            energy_to_electrolyzer_kw = np.ones(8760) * electrolyzer_size_mw * 1e3
    # IF NOT GRID CONNECTED
    else:
        hydrogen_production_capacity_required_kgphr = []
        grid_connection_scenario = "off-grid"
        energy_to_electrolyzer_kw = np.asarray(
            hopp_results["combined_hybrid_power_production_hopp"]
        )
        
    n_pem_clusters = int(ceildiv(electrolyzer_size_mw, greenheart_config["electrolyzer"]["cluster_rating_MW"]))

    ## run using greensteel model
    pem_param_dict = {
        "eol_eff_percent_loss": greenheart_config["electrolyzer"][
            "eol_eff_percent_loss"
        ],
        "uptime_hours_until_eol": greenheart_config["electrolyzer"][
            "uptime_hours_until_eol"
        ],
        "include_degradation_penalty": greenheart_config["electrolyzer"][
            "include_degradation_penalty"
        ],
        "turndown_ratio": greenheart_config["electrolyzer"]["turndown_ratio"],
    }

    if "time_between_replacement" in greenheart_config['electrolyzer']:
        warnings.warn("`time_between_replacement` as an input is deprecated. It is now calculated internally and is output in electrolyzer_physics_results['H2_Results']['Time Until Replacement [hrs]'].")

    H2_Results, h2_ts, h2_tot, power_to_electrolyzer_kw = run_h2_PEM(
        electrical_generation_timeseries=energy_to_electrolyzer_kw,
        electrolyzer_size=electrolyzer_size_mw,
        useful_life=greenheart_config["project_parameters"][
            "project_lifetime"
        ],  # EG: should be in years for full plant life - only used in financial model
        n_pem_clusters=n_pem_clusters,
        pem_control_type=greenheart_config["electrolyzer"]["pem_control_type"],
        electrolyzer_direct_cost_kw=electrolyzer_capex_kw,
        user_defined_pem_param_dictionary=pem_param_dict,
        grid_connection_scenario=grid_connection_scenario,  # if not offgrid, assumes steady h2 demand in kgphr for full year
        hydrogen_production_capacity_required_kgphr=hydrogen_production_capacity_required_kgphr,
        debug_mode=False,
        verbose=verbose,
    )

    # calculate mass and foorprint of system
    mass_kg = run_electrolyzer_mass(electrolyzer_size_mw)
    footprint_m2 = run_electrolyzer_footprint(electrolyzer_size_mw)

    # store results for return
    electrolyzer_physics_results = {
        "H2_Results": H2_Results,
        "capacity_factor": H2_Results["Life: Capacity Factor"],
        "equipment_mass_kg": mass_kg,
        "equipment_footprint_m2": footprint_m2,
        "power_to_electrolyzer_kw": power_to_electrolyzer_kw,
    }

    if verbose:
        print("\nElectrolyzer Physics:")  # 61837444.34555772 145297297.29729727
        print(
            "H2 Produced Annually (tonnes): ",
            H2_Results["Life: Annual H2 production [kg/year]"] * 1e-3,
        )
        print(
            "Max H2 hourly (tonnes): ",
            max(H2_Results["Hydrogen Hourly Production [kg/hr]"]) * 1e-3,
        )
        print(
            "Max H2 daily (tonnes): ",
            max(
                np.convolve(
                    H2_Results["Hydrogen Hourly Production [kg/hr]"],
                    np.ones(24),
                    mode="valid",
                )
            )
            * 1e-3,
        )

        prodrate = 1.0 / round(
            H2_Results["Rated BOL: Efficiency [kWh/kg]"], 2
        )  # kg/kWh
        roughest = power_to_electrolyzer_kw * prodrate
        print("Energy to electrolyzer (kWh): ", sum(power_to_electrolyzer_kw))
        print(
            "Energy per kg (kWh/kg): ",
            H2_Results["Sim: Total Input Power [kWh]"]
            / H2_Results["Sim: Total H2 Produced [kg]"],
        )
        print("Max hourly based on est kg/kWh (kg): ", max(roughest))
        print(
            "Max daily rough est (tonnes): ",
            max(np.convolve(roughest, np.ones(24), mode="valid")) * 1e-3,
        )
        print(
            "Electrolyzer Life Average Capacity Factor: ",
            H2_Results["Life: Capacity Factor"],
        )

    if save_plots or show_plots:
        N = 24 * 7 * 4
        fig, ax = plt.subplots(3, 2, sharex=True, sharey="row")

        wind_speed = [W[2] for W in wind_resource._data["data"]]

        # plt.title("4-week running average")
        pad = 5
        ax[0, 0].annotate(
            "Hourly",
            xy=(0.5, 1),
            xytext=(0, pad),
            xycoords="axes fraction",
            textcoords="offset points",
            size="large",
            ha="center",
            va="baseline",
        )
        ax[0, 1].annotate(
            "4-week running average",
            xy=(0.5, 1),
            xytext=(0, pad),
            xycoords="axes fraction",
            textcoords="offset points",
            size="large",
            ha="center",
            va="baseline",
        )

        ax[0, 0].plot(wind_speed)
        convolved_wind_speed = np.convolve(wind_speed, np.ones(N) / (N), mode="valid")
        ave_x = range(N, len(convolved_wind_speed) + N)
        
        ax[0, 1].plot(ave_x, convolved_wind_speed)
        ax[0, 0].set(ylabel="Wind\n(m/s)", ylim=[0, 30], xlim=[0, len(wind_speed)])
        tick_spacing = 10
        ax[0, 0].yaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))

        y = greenheart_config["electrolyzer"]["rating"]
        ax[1, 0].plot(energy_to_electrolyzer_kw * 1e-3)
        ax[1, 0].axhline(y=y, color="r", linestyle="--", label="Nameplate Capacity")
        
        convolved_energy_to_electrolyzer = np.convolve(
                energy_to_electrolyzer_kw * 1e-3, np.ones(N) / (N), mode="valid"
            )
        
        ax[1, 1].plot(
            ave_x,
            convolved_energy_to_electrolyzer,
        )
        ax[1, 1].axhline(y=y, color="r", linestyle="--", label="Nameplate Capacity")
        ax[1, 0].set(
            ylabel="Electrolyzer \nPower (MW)", ylim=[0, 500], xlim=[0, len(wind_speed)]
        )
        # ax[1].legend(frameon=False, loc="best")
        tick_spacing = 200
        ax[1, 0].yaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))
        ax[1, 0].text(1000, y + 0.1 * tick_spacing, "Electrolyzer Rating", color="r")

        ax[2, 0].plot(
            electrolyzer_physics_results["H2_Results"][
                "Hydrogen Hourly Production [kg/hr]"
            ]
            * 1e-3
        )
        convolved_hydrogen_production = np.convolve(electrolyzer_physics_results["H2_Results"]["Hydrogen Hourly Production [kg/hr]"]*1e-3,np.ones(N) / (N), mode="valid")
        ax[2, 1].plot(
            ave_x,
            convolved_hydrogen_production,
        )
        tick_spacing = 2
        ax[2, 0].set(
            xlabel="Hour",
            ylabel="Hydrogen\n(tonnes/hr)",
            # ylim=[0, 7000],
            xlim=[0, len(H2_Results["Hydrogen Hourly Production [kg/hr]"])],
        )
        ax[2, 0].yaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))

        ax[2, 1].set(
            xlabel="Hour",
            # ylim=[0, 7000],
            xlim=[
                4 * 7 * 24 - 1,
                len(H2_Results["Hydrogen Hourly Production [kg/hr]"] + 4 * 7 * 24 + 2),
            ],
        )
        ax[2, 1].yaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))

        plt.tight_layout()
        if save_plots:
            savepaths = [
                output_dir + "figures/production/",
                output_dir + "data/",
            ]
            for savepath in savepaths:
                if not os.path.exists(savepath):
                    os.makedirs(savepath)
            plt.savefig(
                savepaths[0] + "production_overview_%i.png" % (design_scenario["id"]),
                transparent=True,
            )
            pd.DataFrame.from_dict(
                data={
                    "Hydrogen Hourly Production [kg/hr]": H2_Results[
                        "Hydrogen Hourly Production [kg/hr]"
                    ],
                    "Hourly Water Consumption [kg/hr]": electrolyzer_physics_results[
                        "H2_Results"
                    ]["Water Hourly Consumption [kg/hr]"],
                }
            ).to_csv(savepaths[1] + "h2_flow_%i.csv" % (design_scenario["id"]))
        if show_plots:
            plt.show()

    return electrolyzer_physics_results


def run_electrolyzer_cost(
    electrolyzer_physics_results,
    hopp_config,
    greenheart_config,
    design_scenario,
    verbose=False,
):

    # unpack inputs
    H2_Results = electrolyzer_physics_results["H2_Results"]
    electrolyzer_size_mw = greenheart_config["electrolyzer"]["rating"]
    useful_life = greenheart_config["project_parameters"]["project_lifetime"]
    atb_year = greenheart_config["project_parameters"]["atb_year"]
    electrical_generation_timeseries = electrolyzer_physics_results[
        "power_to_electrolyzer_kw"
    ]
    nturbines = hopp_config["technologies"]["wind"]["num_turbines"]

    electrolyzer_cost_model = greenheart_config["electrolyzer"][
        "cost_model"
    ]  # can be "basic" or "singlitico2021"

    # run hydrogen production cost model - from hopp examples
    if design_scenario["electrolyzer_location"] == "onshore":
        offshore = 0
    else:
        offshore = 1

    if design_scenario["electrolyzer_location"] == "turbine":
        per_turb_electrolyzer_size_mw = electrolyzer_size_mw / nturbines
        per_turb_h2_annual_output = H2_Results["hydrogen_annual_output"] / nturbines
        per_turb_electrical_generation_timeseries = (
            electrical_generation_timeseries / nturbines
        )

        if electrolyzer_cost_model == "basic":
            (
                cf_h2_annuals,
                per_turb_electrolyzer_total_capital_cost,
                per_turb_electrolyzer_OM_cost,
                per_turb_electrolyzer_capex_kw,
                time_between_replacement,
                h2_tax_credit,
                h2_itc,
            ) = basic_H2_cost_model(
                greenheart_config["electrolyzer"]["electrolyzer_capex"],
                greenheart_config["electrolyzer"]["time_between_replacement"],
                per_turb_electrolyzer_size_mw,
                useful_life,
                atb_year,
                per_turb_electrical_generation_timeseries,
                per_turb_h2_annual_output,
                0.0,
                0.0,
                include_refurb_in_opex=False,
                offshore=offshore,
            )

        elif electrolyzer_cost_model == "singlitico2021":

            P_elec = per_turb_electrolyzer_size_mw * 1e-3  # [GW]
            RC_elec = greenheart_config["electrolyzer"][
                "electrolyzer_capex"
            ]  # [USD/kW]

            pem_offshore = PEMCostsSingliticoModel(elec_location=offshore)

            (
                per_turb_electrolyzer_capital_cost_musd,
                per_turb_electrolyzer_om_cost_musd,
            ) = pem_offshore.run(P_elec, RC_elec)

            per_turb_electrolyzer_total_capital_cost = (
                per_turb_electrolyzer_capital_cost_musd * 1e6
            )  # convert from M USD to USD
            per_turb_electrolyzer_OM_cost = (
                per_turb_electrolyzer_om_cost_musd * 1e6
            )  # convert from M USD to USD

        electrolyzer_total_capital_cost = (
            per_turb_electrolyzer_total_capital_cost * nturbines
        )
        electrolyzer_OM_cost = per_turb_electrolyzer_OM_cost * nturbines

    else:
        if electrolyzer_cost_model == "basic":
            (
                cf_h2_annuals,
                electrolyzer_total_capital_cost,
                electrolyzer_OM_cost,
                electrolyzer_capex_kw,
                time_between_replacement,
                h2_tax_credit,
                h2_itc,
            ) = basic_H2_cost_model(
                greenheart_config["electrolyzer"]["electrolyzer_capex"],
                greenheart_config["electrolyzer"]["time_between_replacement"],
                electrolyzer_size_mw,
                useful_life,
                atb_year,
                electrical_generation_timeseries,
                H2_Results["hydrogen_annual_output"],
                0.0,
                0.0,
                include_refurb_in_opex=False,
                offshore=offshore,
            )
        elif electrolyzer_cost_model == "singlitico2021":

            P_elec = electrolyzer_size_mw * 1e-3  # [GW]
            RC_elec = greenheart_config["electrolyzer"][
                "electrolyzer_capex"
            ]  # [USD/kW]

            pem_offshore = PEMCostsSingliticoModel(elec_location=offshore)

            (
                electrolyzer_capital_cost_musd,
                electrolyzer_om_cost_musd,
            ) = pem_offshore.run(P_elec, RC_elec)

            electrolyzer_total_capital_cost = (
                electrolyzer_capital_cost_musd * 1e6
            )  # convert from M USD to USD
            electrolyzer_OM_cost = (
                electrolyzer_om_cost_musd * 1e6
            )  # convert from M USD to USD

        else:
            raise (
                ValueError(
                    "Electrolyzer cost model must be one of['basic', 'singlitico2021'] but '%s' was given"
                    % (electrolyzer_cost_model)
                )
            )

    # package outputs for return
    electrolyzer_cost_results = {
        "electrolyzer_total_capital_cost": electrolyzer_total_capital_cost,
        "electrolyzer_OM_cost_annual": electrolyzer_OM_cost,
    }

    # print some results if desired
    if verbose:
        print("\nHydrogen Cost Results:")
        print(
            "Electrolyzer Total CAPEX $/kW: ",
            electrolyzer_total_capital_cost / (electrolyzer_size_mw * 1e3),
        )
        print(
            "Electrolyzer O&M $/kW: ",
            electrolyzer_OM_cost / (electrolyzer_size_mw * 1e3),
        )
        print(
            "Electrolyzer O&M $/kg: ",
            electrolyzer_OM_cost
            / electrolyzer_physics_results["H2_Results"][
                "Life: Annual H2 production [kg/year]"
            ],
        )

    return electrolyzer_cost_results

def run_electrolyzer_bop(
        plant_config,
        electrolyzer_physics_results,

):
    if "include_bop_power" in plant_config["electrolyzer"]:
        if plant_config['electrolyzer']['include_bop_power']:
            energy_consumption_bop = pem_bop(electrolyzer_physics_results["power_to_electrolyzer_kw"],
                                            plant_config["electrolyzer"]["rating"],
                                            plant_config["electrolyzer"]["turndown_ratio"])
            warnings.warn(
                "Electrolyzer BOP energy consumption is dominated by power electronics (AC-DC conversion and step-down) if electrical system is different consider setting `include_bop_power` to False.",
                UserWarning
                )
        else:
            energy_consumption_bop = np.zeros(len(electrolyzer_physics_results["power_to_electrolyzer_kw"]))
    else:
        energy_consumption_bop = np.zeros(len(electrolyzer_physics_results["power_to_electrolyzer_kw"]))
    return energy_consumption_bop


def run_desal(
    plant_config, electrolyzer_physics_results, design_scenario, verbose=False
):
    if verbose:
        print("\n")
        print("Desal Results")

    if design_scenario["electrolyzer_location"] == "onshore":
        desal_results = {
            "feed_water_flowrat_m3perhr": 0,
            "desal_capex_usd": 0,
            "desal_opex_usd_per_year": 0,
            "power_for_desal_kw": 0,
            "fresh_water_capacity_m3_per_hour": 0,
            "equipment_mass_kg": 0,
            "equipment_footprint_m2": 0,
        }
    else:
        freshwater_kg_per_hr = np.mean(
            electrolyzer_physics_results["H2_Results"][
                "Water Hourly Consumption [kg/hr]"
            ]
        )  # to kg/hr

        if design_scenario["electrolyzer_location"] == "platform":
            (
                desal_capacity_m3_per_hour,
                feedwater_m3_per_hr,
                desal_power,
                desal_capex,
                desal_opex,
                desal_mass_kg,
                desal_size_m2,
            ) = RO_desal(freshwater_kg_per_hr, salinity="Seawater")

            # package outputs
            desal_results = {
                "fresh_water_flowrate_m3perhr": desal_capacity_m3_per_hour,
                "feed_water_flowrat_m3perhr": feedwater_m3_per_hr,
                "desal_capex_usd": desal_capex,
                "desal_opex_usd_per_year": desal_opex,
                "power_for_desal_kw": desal_power,
                "equipment_mass_kg": desal_mass_kg,
                "equipment_footprint_m2": desal_size_m2,
            }

            if verbose:
                print("Fresh water needed (m^3/hr): ", desal_capacity_m3_per_hour)

        elif design_scenario["electrolyzer_location"] == "turbine":
            nturbines = plant_config["technologies"]["wind"]["num_turbines"]

            # size for per-turbine desal #TODO consider using individual power generation time series from each turbine
            in_turb_freshwater_kg_per_hr = freshwater_kg_per_hr / nturbines

            (
                per_turb_desal_capacity_m3_per_hour,
                per_turb_feedwater_m3_per_hr,
                per_turb_desal_power,
                per_turb_desal_capex,
                per_turb_desal_opex,
                per_turb_desal_mass_kg,
                per_turb_desal_size_m2,
            ) = RO_desal(in_turb_freshwater_kg_per_hr, salinity="Seawater")

            fresh_water_flowrate = nturbines * per_turb_desal_capacity_m3_per_hour
            feed_water_flowrate = nturbines * per_turb_feedwater_m3_per_hr
            desal_capex = nturbines * per_turb_desal_capex
            desal_opex = nturbines * per_turb_desal_opex
            power_for_desal = nturbines * per_turb_desal_power

            # package outputs
            desal_results = {
                "fresh_water_flowrate_m3perhr": fresh_water_flowrate,
                "feed_water_flowrat_m3perhr": feed_water_flowrate,
                "desal_capex_usd": desal_capex,
                "desal_opex_usd_per_year": desal_opex,
                "power_for_desal_kw": power_for_desal,
                "per_turb_equipment_mass_kg": per_turb_desal_mass_kg,
                "per_turb_equipment_footprint_m2": per_turb_desal_size_m2,
            }

        if verbose:
            print(
                "Fresh water needed (m^3/hr): ",
                desal_results["fresh_water_flowrate_m3perhr"],
            )
            print("Requested fresh water (m^3/hr):", freshwater_kg_per_hr / 997)

    if verbose:
        for key in desal_results.keys():
            print("Average", key, " ", np.average(desal_results[key]))
        print("\n")

    return desal_results



import numpy as np
import pandas as pd
import warnings
import copy

from ORBIT import ProjectManager, load_config
from ORBIT.core import Vessel
from ORBIT.core.library import initialize_library
from ORBIT.phases.design import DesignPhase
from ORBIT.phases.install import InstallPhase

from greenheart.simulation.technologies.hydrogen.h2_transport.h2_compression import (
    Compressor,
)
from greenheart.simulation.technologies.hydrogen.h2_storage.pressure_vessel.compressed_gas_storage_model_20221021.Compressed_all import (
    PressureVessel,
)
from greenheart.simulation.technologies.hydrogen.h2_storage.pipe_storage import (
    UndergroundPipeStorage,
)

from greenheart.simulation.technologies.hydrogen.h2_storage.lined_rock_cavern.lined_rock_cavern import (
    LinedRockCavernStorage,
)
from greenheart.simulation.technologies.hydrogen.h2_storage.salt_cavern.salt_cavern import (
    SaltCavernStorage,
)
from greenheart.simulation.technologies.hydrogen.h2_storage.on_turbine.on_turbine_hydrogen_storage import (
    PressurizedTower,
)

from greenheart.simulation.technologies.hydrogen.h2_transport.h2_export_pipe import (
    run_pipe_analysis,
)
from greenheart.simulation.technologies.hydrogen.h2_transport.h2_pipe_array import (
    run_pipe_array_const_diam,
)
from greenheart.simulation.technologies.offshore.fixed_platform import (
    FixedPlatformDesign,
    FixedPlatformInstallation,
)
from greenheart.simulation.technologies.offshore.floating_platform import (
    FloatingPlatformDesign,
    FloatingPlatformInstallation,
)
from greenheart.simulation.technologies.offshore.all_platforms import calc_platform_opex

from greenheart.simulation.technologies.hydrogen.h2_storage.storage_sizing import (
    hydrogen_storage_capacity,
)


def run_h2_pipe_array(
    greenheart_config,
    hopp_config,
    turbine_config,
    wind_cost_results,
    electrolyzer_physics_results,
    design_scenario,
    verbose,
):
    if design_scenario["transportation"] == "hvdc+pipeline" or (
        design_scenario["electrolyzer_location"] == "turbine"
        and not design_scenario["h2_storage_location"] == "turbine"
    ):
        # get pipe lengths from ORBIT using cable lengths (horizontal only)
        if design_scenario["wind_location"] == "offshore":
            pipe_lengths = wind_cost_results.orbit_project.phases[
                "ArraySystemDesign"
            ].sections_distance

        if (
            design_scenario["wind_location"] == "onshore"
        ):  # TODO: improve pipe length estimate
            raise NotImplementedError(
                "distributed electrolysis is not implemented for onshore wind."
            )
            # np.ones_like(hopp_config["technologies"]["wind"]["num_turbines"])*(
            #     greenheart_config["site"]["wind_layout"]["turbine_spacing"]
            #     *turbine_config['rotor_diameter']
            # )

        turbine_h2_flowrate = (
            max(
                electrolyzer_physics_results["H2_Results"][
                    "Hydrogen Hourly Production [kg/hr]"
                ]
            )
            * ((1.0 / 60.0) ** 2)
            / hopp_config["technologies"]["wind"]["num_turbines"]
        )
        m_dot = (
            np.ones_like(pipe_lengths) * turbine_h2_flowrate
        )  # Mass flow rate [kg/s] assuming 300 MW -> 1.5 kg/s
        p_inlet = (
            31  # Inlet pressure [bar] - assumed outlet pressure from electrolyzer model
        )
        p_outlet = 10  # Outlet pressure [bar] - about 20 bar drop
        depth = greenheart_config["site"]["depth"]  # depth of pipe [m]

        capex, opex = run_pipe_array_const_diam(
            pipe_lengths, depth, p_inlet, p_outlet, m_dot
        )

        h2_pipe_array_results = {"capex": capex, "opex": opex}
    else:
        h2_pipe_array_results = {"capex": 0.0, "opex": 0.0}

    return h2_pipe_array_results


def run_h2_transport_compressor(
    greenheart_config, electrolyzer_physics_results, design_scenario, verbose=False
):
    if (
        design_scenario["transportation"] == "pipeline"
        or design_scenario["transportation"] == "hvdc+pipeline"
        or (
            design_scenario["h2_storage_location"] != "onshore"
            and design_scenario["electrolyzer_location"] == "onshore"
        )
    ):
        ########## compressor model from Jamie Kee based on HDSAM
        flow_rate_kg_per_hr = max(
            electrolyzer_physics_results["H2_Results"][
                "Hydrogen Hourly Production [kg/hr]"
            ]
        )  # kg/hr
        number_of_compressors = 2  # a third will be added as backup in the code
        p_inlet = 20  # bar
        p_outlet = greenheart_config["h2_transport_compressor"][
            "outlet_pressure"
        ]  # bar
        flow_rate_kg_d = flow_rate_kg_per_hr * 24.0

        compressor = Compressor(
            p_outlet,
            flow_rate_kg_d,
            p_inlet=p_inlet,
            n_compressors=number_of_compressors,
        )
        compressor.compressor_power()
        motor_rating, system_power_kw = compressor.compressor_system_power()
        total_capex, total_OM = compressor.compressor_costs()  # 2016$ , 2016$/y

        # print(f"CAPEX: {round(total_capex,2)} $")
        # print(f"Annual operating expense: {round(total_OM,2)} $/yr")

        h2_transport_compressor_results = {
            "compressor_power": system_power_kw,
            "compressor_capex": total_capex,
            "compressor_opex": total_OM,
        }

    else:
        compressor = None
        h2_transport_compressor_results = {
            "compressor_power": 0.0,
            "compressor_capex": 0.0,
            "compressor_opex": 0.0,
        }
        flow_rate_kg_per_hr = 0.0

    if verbose:
        print("\nCompressor Results:")
        print("Total H2 Flowrate (kg/hr): ", flow_rate_kg_per_hr)
        print(
            "Compressor_power (kW): ",
            h2_transport_compressor_results["compressor_power"],
        )
        print(
            "Compressor capex [USD]: ",
            h2_transport_compressor_results["compressor_capex"],
        )
        print(
            "Compressor opex [USD/yr]: ",
            h2_transport_compressor_results["compressor_opex"],
        )  # annual

    return compressor, h2_transport_compressor_results


def run_h2_transport_pipe(
    orbit_config,
    greenheart_config,
    electrolyzer_physics_results,
    design_scenario,
    verbose=False,
):
    # prepare inputs
    export_pipe_length = orbit_config["site"]["distance_to_landfall"]  # Length [km]
    mass_flow_rate = max(
        electrolyzer_physics_results["H2_Results"]["Hydrogen Hourly Production [kg/hr]"]
    ) * (
        (1.0 / 60.0) ** 2
    )  # from [kg/hr] to mass flow rate in [kg/s] assuming 300 MW -> 1.5 kg/s
    p_inlet = greenheart_config["h2_transport_compressor"][
        "outlet_pressure"
    ]  # Inlet pressure [bar]
    p_outlet = greenheart_config["h2_transport_pipe"][
        "outlet_pressure"
    ]  # Outlet pressure [bar]
    depth = greenheart_config["site"]["depth"]  # depth of pipe [m]

    # run model
    if (
        design_scenario["transportation"] == "pipeline"
        or design_scenario["transportation"] == "hvdc+pipeline"
    ) or (
        design_scenario["h2_storage_location"] != "onshore"
        and design_scenario["electrolyzer_location"] == "onshore"
    ):
        h2_transport_pipe_results = run_pipe_analysis(
            export_pipe_length, mass_flow_rate, p_inlet, p_outlet, depth
        )
    else:
        h2_transport_pipe_results = pd.DataFrame.from_dict(
            {
                "index": 0,
                "Grade": ["none"],
                "Outer diameter (mm)": [0 * 141.3],
                "Inner Diameter (mm)": [0 * 134.5],
                "Schedule": ["none"],
                "Thickness (mm)": [0 * 3.4],
                "volume [m3]": [0 * 30.969133941093407],
                "weight [kg]": [0 * 242798.01009817232],
                "mat cost [$]": [0 * 534155.6222159792],
                "labor cost [$]": [0 * 2974375.749734022],
                "misc cost [$]": [0 * 970181.8962542458],
                "ROW cost [$]": [0 * 954576.9166912301],
                "total capital cost [$]": [0 * 5433290.0184895478],
                "annual operating cost [$]": [0.0],
            }
        )
    if verbose:
        print("\nH2 Transport Pipe Results")
        for col in h2_transport_pipe_results.columns:
            if col == "index":
                continue
            print(col, h2_transport_pipe_results[col][0])
        print("\n")

    return h2_transport_pipe_results


def run_h2_storage(
    hopp_config,
    greenheart_config,
    turbine_config,
    electrolyzer_physics_results,
    design_scenario,
    verbose=False,
):

    if design_scenario["h2_storage_location"] == "platform":
        if (
            greenheart_config["h2_storage"]["type"] != "pressure_vessel"
            and greenheart_config["h2_storage"]["type"] != "none"
        ):
            raise ValueError(
                "Only pressure vessel storage can be used on the off shore platform"
            )

    if design_scenario["h2_storage_location"] == "turbine":
        if (
            greenheart_config["h2_storage"]["type"] != "turbine"
            and greenheart_config["h2_storage"]["type"] != "pressure_vessel"
            and greenheart_config["h2_storage"]["type"] != "none"
        ):
            raise ValueError(
                "Only turbine or pressure vessel storage can be used for turbine hydrogen storage location"
            )
    ########### initialize output dictionary ###########
    h2_storage_results = dict()

    storage_max_fill_rate = np.max(
        electrolyzer_physics_results["H2_Results"]["Hydrogen Hourly Production [kg/hr]"]
    )

    ########### get hydrogen storage size in kilograms ###########
    ##################### no hydrogen storage
    if greenheart_config["h2_storage"]["type"] == "none":
        h2_storage_capacity_kg = 0.0
        storage_max_fill_rate = 0.0

    ##################### get storage capacity from turbine storage model
    elif greenheart_config["h2_storage"]["capacity_from_max_on_turbine_storage"]:
        nturbines = hopp_config["technologies"]["wind"]["num_turbines"]
        turbine = {
            "tower_length": turbine_config["tower"]["length"],
            "section_diameters": turbine_config["tower"]["section_diameters"],
            "section_heights": turbine_config["tower"]["section_heights"],
        }

        h2_storage = PressurizedTower(
            greenheart_config["project_parameters"]["atb_year"], turbine
        )
        h2_storage.run()

        h2_storage_capacity_single_turbine = h2_storage.get_capacity_H2()  # kg

        h2_storage_capacity_kg = nturbines * h2_storage_capacity_single_turbine  # in kg

    ##################### get storage capacity from hydrogen storage demand
    elif greenheart_config["h2_storage"]["size_capacity_from_demand"]["flag"]:
        hydrogen_storage_demand = np.mean(
            electrolyzer_physics_results["H2_Results"][
                "Hydrogen Hourly Production [kg/hr]"
            ]
        )  # TODO: update demand based on end-use needs
        (
            hydrogen_storage_capacity_kg,
            hydrogen_storage_duration_hr,
            hydrogen_storage_soc,
        ) = hydrogen_storage_capacity(
            electrolyzer_physics_results["H2_Results"],
            greenheart_config["electrolyzer"]["rating"],
            hydrogen_storage_demand,
        )
        h2_storage_capacity_kg = hydrogen_storage_capacity_kg
        h2_storage_results["hydrogen_storage_duration_hr"] = (
            hydrogen_storage_duration_hr
        )
        h2_storage_results["hydrogen_storage_soc"] = hydrogen_storage_soc

    ##################### get storage capacity based on storage days in config
    else:
        storage_hours = greenheart_config["h2_storage"]["days"] * 24
        h2_storage_capacity_kg = round(storage_hours * storage_max_fill_rate)

    h2_storage_results["h2_storage_capacity_kg"] = h2_storage_capacity_kg
    h2_storage_results["h2_storage_max_fill_rate_kg_hr"] = storage_max_fill_rate

    ########### run specific hydrogen storage models for costs and energy use ###########
    if (
        greenheart_config["h2_storage"]["type"] == "none"
        or design_scenario["h2_storage_location"] == "none"
    ):
        h2_storage_results["storage_capex"] = 0.0
        h2_storage_results["storage_opex"] = 0.0
        h2_storage_results["storage_energy"] = 0.0

        h2_storage = None

    elif greenheart_config["h2_storage"]["type"] == "turbine":
        if design_scenario["h2_storage_location"] == "turbine":
            turbine = {
                "tower_length": turbine_config["tower"]["length"],
                "section_diameters": turbine_config["tower"]["section_diameters"],
                "section_heights": turbine_config["tower"]["section_heights"],
            }

            h2_storage = PressurizedTower(
                greenheart_config["project_parameters"]["atb_year"], turbine
            )
            h2_storage.run()

            h2_storage_results["storage_capex"] = nturbines * h2_storage.get_capex()
            h2_storage_results["storage_opex"] = nturbines * h2_storage.get_opex()

            if verbose:
                print("On-turbine H2 storage:")
                print("mass empty (single turbine): ", h2_storage.get_mass_empty())
                print(
                    "H2 capacity (kg) - single turbine: ", h2_storage.get_capacity_H2()
                )
                print("storage pressure: ", h2_storage.get_pressure_H2())

            h2_storage_results["storage_energy"] = (
                0.0  # low pressure, so no additional compression needed beyond electolyzer
            )
        else:
            raise ValueError(
                "`h2_storage_location` must be set to 'turbine' to use 'turbine' for h2 storage type."
            )

    elif greenheart_config["h2_storage"]["type"] == "pipe":
        # for more information, see https://www.nrel.gov/docs/fy14osti/58564.pdf
        # initialize dictionary for pipe storage parameters
        storage_input = dict()

        # pull parameters from plat_config file
        storage_input["h2_storage_kg"] = h2_storage_capacity_kg
        storage_input["compressor_output_pressure"] = greenheart_config[
            "h2_storage_compressor"
        ]["output_pressure"]
        storage_input["system_flow_rate"] = storage_max_fill_rate
        storage_input["model"] = "papadias"

        # run pipe storage model
        h2_storage = UndergroundPipeStorage(storage_input)

        h2_storage.pipe_storage_capex()
        h2_storage.pipe_storage_opex()

        h2_storage_results["storage_capex"] = h2_storage.output_dict[
            "pipe_storage_capex"
        ]
        h2_storage_results["storage_opex"] = h2_storage.output_dict["pipe_storage_opex"]
        h2_storage_results["storage_energy"] = 0.0

    elif greenheart_config["h2_storage"]["type"] == "pressure_vessel":
        if design_scenario["h2_storage_location"] == "turbine":
            energy_cost = 0.0

            h2_storage = PressureVessel(Energy_cost=energy_cost)
            h2_storage.run()

            (
                capex_dist_total,
                opex_dist_total,
                energy,
                area_site,
                mass_tank_empty_site,
                _,
            ) = h2_storage.distributed_storage_vessels(h2_storage_capacity_kg, 1)
            # ) = h2_storage.distributed_storage_vessels(h2_capacity, nturbines)
            # capex, opex, energy = h2_storage.calculate_from_fit(h2_capacity)

            h2_storage_results["storage_capex"] = capex_dist_total
            h2_storage_results["storage_opex"] = opex_dist_total
            h2_storage_results["storage_energy"] = (
                energy
                * electrolyzer_physics_results["H2_Results"][
                    "Life: Annual H2 production [kg/year]"
                ]
            )  # total in kWh
            h2_storage_results["tank_mass_full_kg"] = (
                h2_storage.get_tank_mass(h2_storage_capacity_kg)[1]
                + h2_storage_capacity_kg
            )
            h2_storage_results["tank_footprint_m2"] = h2_storage.get_tank_footprint(
                h2_storage_capacity_kg, upright=True
            )[1]
            h2_storage_results["tank volume (m^3)"] = (
                h2_storage.compressed_gas_function.Vtank
            )
            h2_storage_results["Number of tanks"] = h2_storage.get_tanks(
                h2_storage_capacity_kg
            )
            if verbose:
                print("ENERGY FOR STORAGE: ", energy * 1e-3 / (365 * 24), " MW")
                print("Tank volume (M^3): ", h2_storage_results["tank volume (m^3)"])
                print(
                    "Single Tank capacity (kg): ",
                    h2_storage.compressed_gas_function.single_tank_h2_capacity_kg,
                )
                print("N Tanks: ", h2_storage_results["Number of tanks"])

        else:
            # if plant_config["project_parameters"]["grid_connection"]:
            #     energy_cost = plant_config["project_parameters"]["ppa_price"]
            # else:
            #     energy_cost = 0.0
            energy_cost = 0.0  # energy cost is now handled outside the storage model

            h2_storage = PressureVessel(Energy_cost=energy_cost)
            h2_storage.run()

            capex, opex, energy = h2_storage.calculate_from_fit(h2_storage_capacity_kg)

            h2_storage_results["storage_capex"] = capex
            h2_storage_results["storage_opex"] = opex
            h2_storage_results["storage_energy"] = (
                energy
                * electrolyzer_physics_results["H2_Results"][
                    "Life: Annual H2 production [kg/year]"
                ]
            )  # total in kWh
            h2_storage_results["tank_mass_full_kg"] = (
                h2_storage.get_tank_mass(h2_storage_capacity_kg)[1]
                + h2_storage_capacity_kg
            )
            h2_storage_results["tank_footprint_m2"] = h2_storage.get_tank_footprint(
                h2_storage_capacity_kg, upright=True
            )[1]
            h2_storage_results["tank volume (m^3)"] = (
                h2_storage.compressed_gas_function.Vtank
            )
            h2_storage_results["Number of tanks"] = (
                h2_storage.compressed_gas_function.number_of_tanks
            )
            if verbose:
                print("ENERGY FOR STORAGE: ", energy * 1e-3 / (365 * 24), " MW")
                print("Tank volume (M^3): ", h2_storage_results["tank volume (m^3)"])
                print(
                    "Single Tank capacity (kg): ",
                    h2_storage.compressed_gas_function.single_tank_h2_capacity_kg,
                )
                print("N Tanks: ", h2_storage_results["Number of tanks"])

    elif greenheart_config["h2_storage"]["type"] == "salt_cavern":
        # initialize dictionary for salt cavern storage parameters
        storage_input = dict()

        # pull parameters from plant_config file
        storage_input["h2_storage_kg"] = h2_storage_capacity_kg
        storage_input["system_flow_rate"] = storage_max_fill_rate
        storage_input["model"] = "papadias"

        # run salt cavern storage model
        h2_storage = SaltCavernStorage(storage_input)

        h2_storage.salt_cavern_capex()
        h2_storage.salt_cavern_opex()

        h2_storage_results["storage_capex"] = h2_storage.output_dict[
            "salt_cavern_storage_capex"
        ]
        h2_storage_results["storage_opex"] = h2_storage.output_dict[
            "salt_cavern_storage_opex"
        ]
        h2_storage_results["storage_energy"] = 0.0

    elif greenheart_config["h2_storage"]["type"] == "lined_rock_cavern":
        # initialize dictionary for salt cavern storage parameters
        storage_input = dict()

        # pull parameters from plat_config file
        storage_input["h2_storage_kg"] = h2_storage_capacity_kg
        storage_input["system_flow_rate"] = storage_max_fill_rate
        storage_input["model"] = "papadias"

        # run salt cavern storage model
        h2_storage = LinedRockCavernStorage(storage_input)

        h2_storage.lined_rock_cavern_capex()
        h2_storage.lined_rock_cavern_opex()

        h2_storage_results["storage_capex"] = h2_storage.output_dict[
            "lined_rock_cavern_storage_capex"
        ]
        h2_storage_results["storage_opex"] = h2_storage.output_dict[
            "lined_rock_cavern_storage_opex"
        ]
        h2_storage_results["storage_energy"] = 0.0
    else:
        raise (
            ValueError(
                "H2 storage type %s was given, but must be one of ['none', 'turbine', 'pipe', 'pressure_vessel', 'salt_cavern', 'lined_rock_cavern']"
            )
        )

    if verbose:
        print("\nH2 Storage Results:")
        print("H2 storage capex: ${0:,.0f}".format(h2_storage_results["storage_capex"]))
        print(
            "H2 storage annual opex: ${0:,.0f}/yr".format(
                h2_storage_results["storage_opex"]
            )
        )
        print(
            "H2 storage capacity (tonnes): ",
            h2_storage_results["h2_storage_capacity_kg"] / 1000,
        )
        if h2_storage_results["h2_storage_capacity_kg"] > 0:
            print(
                "H2 storage cost $/kg of H2: ",
                h2_storage_results["storage_capex"]
                / h2_storage_results["h2_storage_capacity_kg"],
            )

    return h2_storage, h2_storage_results


def run_equipment_platform(
    hopp_config,
    greenheart_config,
    orbit_config,
    design_scenario,
    hopp_results,
    electrolyzer_physics_results,
    h2_storage_results,
    desal_results,
    verbose=False,
):
    topmass = 0.0  # tonnes
    toparea = 0.0  # m^2

    if (
        design_scenario["electrolyzer_location"] == "platform"
        or design_scenario["h2_storage_location"] == "platform"
        or design_scenario["pv_location"] == "platform"
        or design_scenario["battery_location"] == "platform"
    ):
        """ "equipment_mass_kg": desal_mass_kg,
        "equipment_footprint_m2": desal_size_m2"""

        if design_scenario["electrolyzer_location"] == "platform":
            topmass += (
                electrolyzer_physics_results["equipment_mass_kg"] * 1e-3
            )  # from kg to tonnes
            topmass += desal_results["equipment_mass_kg"] * 1e-3  # from kg to tonnes
            toparea += electrolyzer_physics_results["equipment_footprint_m2"]
            toparea += desal_results["equipment_footprint_m2"]

        if (
            design_scenario["h2_storage_location"] == "platform"
            and greenheart_config["h2_storage"]["type"] != "none"
        ):
            topmass += (
                h2_storage_results["tank_mass_full_kg"] * 1e-3
            )  # from kg to tonnes
            toparea += h2_storage_results["tank_footprint_m2"]

        if (
            "battery" in hopp_config["technologies"].keys()
            and design_scenario["battery_location"] == "platform"
        ):
            battery_area = hopp_results["hybrid_plant"].battery.footprint_area
            battery_mass = hopp_results["hybrid_plant"].battery.system_mass

            topmass += battery_mass
            toparea += battery_area

        if (
            hopp_config["site"]["solar"]
            and design_scenario["pv_location"] == "platform"
        ):
            pv_area = hopp_results["hybrid_plant"].pv.footprint_area
            solar_mass = hopp_results["hybrid_plant"].pv.system_mass

            if pv_area > toparea:
                warnings.warn(
                    f"Solar area ({pv_area} m^2) must be smaller than platform area ({toparea} m^2)",
                    UserWarning,
                )
            topmass += solar_mass

        #### initialize
        if (
            greenheart_config["platform"]["design_phases"][0]
            == "FloatingPlatformDesign"
        ):
            if not ProjectManager.find_key_match("FloatingPlatformDesign"):
                ProjectManager.register_design_phase(FloatingPlatformDesign)
            if not ProjectManager.find_key_match("FloatingPlatformInstallation"):
                ProjectManager.register_install_phase(FloatingPlatformInstallation)
        else:
            if not ProjectManager.find_key_match("FixedPlatformDesign"):
                ProjectManager.register_design_phase(FixedPlatformDesign)
            if not ProjectManager.find_key_match("FixedPlatformInstallation"):
                ProjectManager.register_install_phase(FixedPlatformInstallation)

        platform_config = copy.deepcopy(greenheart_config["platform"])

        # assign site parameters
        if platform_config["site"]["depth"] == -1:
            platform_config["site"]["depth"] = greenheart_config["site"]["depth"]
        if platform_config["site"]["distance"] == -1:
            platform_config["site"]["distance"] = orbit_config["site"]["distance"]
        # assign equipment values
        
        if platform_config["equipment"]["tech_combined_mass"] == -1:
            platform_config["equipment"]["tech_combined_mass"] = topmass
        if platform_config["equipment"]["tech_required_area"] == -1:
            platform_config["equipment"]["tech_required_area"] = toparea
        platform = ProjectManager(platform_config)
        platform.run()

        design_capex = platform.design_results["platform_design"]["total_cost"]
        install_capex = platform.installation_capex
        total_capex = design_capex + install_capex

        total_opex = calc_platform_opex(total_capex, platform_config["opex_rate"])

        platform_mass = platform.design_results["platform_design"]["mass"]
        platform_area = platform.design_results["platform_design"]["area"]

    else:
        platform_mass = 0.0
        platform_area = 0.0
        total_capex = 0.0
        total_opex = 0.0

    platform_results = {
        "topmass_kg": topmass,
        "toparea_m2": platform_area,
        "platform_mass_tonnes": platform_mass,
        "capex": total_capex,
        "opex": total_opex,
    }
    
    if verbose:
        print("\nPlatform Results")
        for key in platform_results.keys():
            print(key, "%.2f" % (platform_results[key]))

    return platform_results



"""
Pytest configuration file.
"""
import os

from hopp import TEST_ENV_VAR
from hopp.utilities.keys import set_nrel_key_dot_env

def pytest_sessionstart(session):
    os.environ["ENV"] = TEST_ENV_VAR

    # Set a dummy API key
    os.environ["NREL_API_KEY"] = "a" * 40
    set_nrel_key_dot_env()




"""
Author: Cory Frontin
Date: 23 Jan 2023
Institution: National Renewable Energy Lab
Description: This file handles the cost, sizing, and pressure of on-turbine H2 storage

To use this class, specify a turbine 

Costs are assumed to be in 2003 dollars [1]

Sources:
    - [1] Kottenstette 2003 (use their chosen favorite design)
Args:
    - year (int): construction year
    - turbine (dict): contains various information about the turbine, including tower_length, section_diameters, and section_heights
API member functions:
    - get_capex(): return the total additional capex necessary for H2 production, in 2003 dollars
    - get_opex(): return the result of a simple model for operational expenditures for pressure vessel, in 2003 dollars
    - get_mass_empty(): return the total additional empty mass necessary for H2 production, in kg
    - get_capacity_H2(): return the capacity mass of hydrogen @ operating pressure, ambient temp., in kg
    - get_pressure_H2() return the operating hydrogen pressure, in Pa
"""

import numpy as np

class PressurizedTower():
    def __init__(self,
                 year: int,
                 turbine: dict):
        
        # key inputs
        self.year= year
        self.turbine= turbine
        
        self.tower_length= turbine['tower_length'] # m
        self.section_diameters= turbine['section_diameters'] # m
        self.section_heights= turbine['section_heights'] # m

        # calculation settings
        self.setting_volume_thickness_calc= 'centered' # ['centered', 'outer', 'inner']

        # constants/parameters
        self.d_t_ratio= 320. # Kottenstette 2003
        self.thickness_top= self.section_diameters[-1]/self.d_t_ratio # m
        self.thickness_bot= self.section_diameters[0]/self.d_t_ratio # m
        self.ultimate_tensile_strength= 636e6 # Pa, Kottenstette 2003
        self.yield_strength= 350e6 # Pa, Kottenstette 2003
        self.welded_joint_efficiency= 0.80 # journal edition
        # self.welded_joint_efficiency= 0.85 # double-welded butt joint w/ spot inspection (ASME)
        self.density_steel= 7817. # kg/m^3
        self.gasconstant_H2= 4126. # J/(kg K)
        self.operating_temp= 25. # degC

        self.costrate_steel= 1.50 # $/kg
        self.costrate_endcap= 2.66 # $/kg

        self.costrate_ladder= 32.80 # $/m
        self.cost_door= 2000 # $
        self.cost_mainframe_extension= 6300 # $
        self.cost_nozzles_manway= 16000 # $
        self.costrate_conduit= 35 # $/m

        # based on pressure_vessel maintenance costs
        self.wage= 36 # 2003 dollars (per hour worked)
        self.staff_hours= 60 # hours
        self.maintenance_rate= 0.03 # factor

        # set the operating pressure @ the crossover pressure
        self.operating_pressure= PressurizedTower.get_crossover_pressure(self.welded_joint_efficiency,
                                                                         self.ultimate_tensile_strength,
                                                                         self.d_t_ratio)

    def run(self):

        # get the inner volume and traditional material volume, mass, cost
        self.tower_inner_volume= self.get_volume_tower_inner()
        self.wall_material_volume_trad, self.cap_bot_material_volume_trad, self.cap_top_material_volume_trad= \
                self.get_volume_tower_material(pressure= 0.0)
        self.wall_material_mass_trad= self.wall_material_volume_trad*self.density_steel
        self.wall_material_cost_trad= self.wall_material_mass_trad*self.costrate_steel
        self.cap_material_mass_trad= (self.cap_top_material_volume_trad + self.cap_bot_material_volume_trad)*self.density_steel
        self.cap_material_cost_trad= self.cap_material_mass_trad*self.costrate_steel
        self.nonwall_cost_trad= self.get_cost_nontower(traditional= True)

        self.wall_material_volume, self.cap_bot_material_volume, self.cap_top_material_volume= \
                self.get_volume_tower_material()
        self.wall_material_mass= self.wall_material_volume*self.density_steel
        self.wall_material_cost= self.wall_material_mass*self.costrate_steel
        self.wall_material_mass= self.wall_material_volume*self.density_steel
        self.cap_material_mass= (self.cap_bot_material_volume + self.cap_top_material_volume)*self.density_steel
        self.cap_material_cost= self.cap_material_mass*self.costrate_endcap
        self.nonwall_cost= self.get_cost_nontower()

        if False:
            # print the inner volume and pressure-free material properties
            print("operating pressure:", self.operating_pressure)
            print("tower inner volume:", self.tower_inner_volume)
            print()
            print("tower wall material volume (non-pressurized):",
                self.wall_material_volume_trad)
            print("tower wall material mass (non-pressurized):",
                self.wall_material_mass_trad)
            print("tower wall material cost (non-pressurized):",
                self.wall_material_cost_trad)
            print("tower cap material volume (non-pressurized):",
                self.cap_top_material_volume_trad + self.cap_bot_material_volume_trad)
            print("tower cap material mass (non-pressurized):",
                self.cap_material_mass_trad)
            print("tower cap material cost (non-pressurized):",
                self.cap_material_cost_trad)
            print("tower total material cost (non-pressurized):",
                self.wall_material_cost_trad + self.cap_material_cost_trad)
            
            # print the changes to the structure
            print()
            print("tower wall material volume (pressurized):", self.wall_material_volume)
            print("tower wall material mass (pressurized):", self.wall_material_mass)
            print("tower wall material cost (pressurized):", self.wall_material_cost)
            print()
            print("tower cap material volume (pressurized):", self.cap_bot_material_volume + self.cap_top_material_volume)
            print("tower cap material mass (pressurized):", self.cap_material_mass)
            print("tower top cap material cost (pressurized):", self.cap_top_material_volume*self.density_steel*self.costrate_endcap)
            print("tower bot cap material cost (pressurized):", self.cap_bot_material_volume*self.density_steel*self.costrate_endcap)
            print("tower cap material cost (pressurized):", self.cap_material_cost)
            print()
            print("operating mass fraction:", self.get_operational_mass_fraction())
            print("nonwall cost (non-pressurized):", self.nonwall_cost_trad)
            print("nonwall cost (pressurized):", self.nonwall_cost)
            print("tower total material cost (pressurized):",
                self.wall_material_cost + self.cap_material_cost)

            print()
            print("delta tower wall material cost:", self.wall_material_cost - self.wall_material_cost_trad)
            print("empty mass:", self.get_mass_empty())
            print("")
            print("capex:", self.get_capex())
            print("opex:", self.get_opex())
            print("capacity (H2):", self.get_capacity_H2())

    def get_volume_tower_inner(self):
        """
        get the inner volume of the tower in m^3

        assume t << d
        """

        # count the sections, all assumed conic frustum
        Nsection= len(self.section_diameters) - 1

        # loop over sections, calclulating volume of each
        vol_section= np.zeros((Nsection,))
        for i_section in range(Nsection):
            diameter_bot= self.section_diameters[i_section] # m
            height_bot= self.section_heights[i_section] # m
            diameter_top= self.section_diameters[i_section + 1] # m
            height_top= self.section_heights[i_section + 1] # m
            dh= np.abs(height_top - height_bot) # height of section, m

            vol_section[i_section]= PressurizedTower.compute_frustum_volume(dh,
                                                                            diameter_bot,
                                                                            diameter_top)

        # total volume: sum of sections
        return np.sum(vol_section) # m^3
    
    def get_volume_tower_material(self,
                                  pressure : float = None):
        """
        get the material volume of the tower in m^3

        if pressurized, use pressure to set thickness increment due to pressurization

        assume t << d

        params:
            - pressure: gauge pressure of H2 (defaults to design op. pressure)
        returns:
            - Vmat_wall: material volume of vertical tower
            - Vmat_bot: material volume of bottom cap (nonzero only if pressurized)
            - Vmat_top: material volume of top cap (nonzero only if pressurized)
        """

        # override pressure iff requested
        if pressure is None: pressure= self.operating_pressure

        # this is the linear constant s.t. delta t ~ alpha * d
        alpha_dtp= PressurizedTower.get_thickness_increment_const(pressure, self.ultimate_tensile_strength) # 

        # loop over the sections of the tower
        Nsection= len(self.section_diameters) - 1
        matvol_section= np.zeros((Nsection,))
        for i_section in range(Nsection):
            d1= self.section_diameters[i_section]
            h1= self.section_heights[i_section]
            d2= self.section_diameters[i_section + 1]
            h2= self.section_heights[i_section + 1]
            
            if self.setting_volume_thickness_calc == 'centered':
                Vouter= PressurizedTower.compute_frustum_volume(h2 - h1,
                                                                d1*(1 + (1/self.d_t_ratio + alpha_dtp)),
                                                                d2*(1 + (1/self.d_t_ratio + alpha_dtp)))
                Vinner= PressurizedTower.compute_frustum_volume(h2 - h1,
                                                                d1*(1 - (1/self.d_t_ratio + alpha_dtp)),
                                                                d2*(1 - (1/self.d_t_ratio + alpha_dtp)))
            elif self.setting_volume_thickness_calc == 'outer':
                Vouter= PressurizedTower.compute_frustum_volume(h2 - h1,
                                                                d1*(1 + 2*(1/self.d_t_ratio + alpha_dtp)),
                                                                d2*(1 + 2*(1/self.d_t_ratio + alpha_dtp)))
                Vinner= PressurizedTower.compute_frustum_volume(h2 - h1, d1, d2)
            elif self.setting_volume_thickness_calc == 'inner':
                Vouter= PressurizedTower.compute_frustum_volume(h2 - h1, d1, d2)
                Vinner= PressurizedTower.compute_frustum_volume(h2 - h1,
                                                                d1*(1 - 2*(1/self.d_t_ratio + alpha_dtp)),
                                                                d2*(1 - 2*(1/self.d_t_ratio + alpha_dtp)))

            matvol_section[i_section]= Vouter - Vinner

        # compute wall volume by summing sections
        Vmat_wall= np.sum(matvol_section) # m^3

        if pressure == 0:
            Vmat_bot= 0.0
            Vmat_top= 0.0
        else:
            # compute caps as well: area by thickness
            Vmat_bot= (np.pi/4*self.section_diameters[0]**2) \
                    *(PressurizedTower.compute_cap_thickness(pressure,
                                                             self.section_diameters[0], # assume first is bottom
                                                             self.yield_strength,
                                                             efficiency_weld= self.welded_joint_efficiency)) # m^3
            Vmat_top= np.pi/4*self.section_diameters[-1]**2 \
                    *(PressurizedTower.compute_cap_thickness(pressure,
                                                             self.section_diameters[-1], # assume last is top
                                                             self.yield_strength,
                                                             efficiency_weld= self.welded_joint_efficiency)) # m^3

        # total material volume
        return (Vmat_wall, Vmat_bot, Vmat_top) # m^3
    
    def get_mass_tower_material(self,
                                pressure : float = None):
        """
        get the material mass of the tower in m^3

        if pressurized, use pressure to set thickness increment due to pressurization

        assume t << d

        params:
            - pressure: gauge pressure of H2 (defaults to design op. pressure)
        returns:
            - Mmat_wall: material mass of vertical tower
            - Mmat_bot: material mass of bottom cap
            - Mmat_top: material mass of top cap
        """

        # pass through to volume calculator, multiplying by steel density
        return [self.density_steel*x for x in self.get_volume_tower_material(pressure)] # kg
    
    def get_cost_tower_material(self,
                                pressure : float = None):
        """
        get the material cost of the tower in m^3

        if pressurized, use pressure to set thickness increment due to pressurization

        assume t << d

        params:
            - pressure: gauge pressure of H2 (defaults to design op. pressure)
        returns:
            - Vmat_wall: material cost of vertical tower
            - Vmat_bot: material cost of bottom cap
            - Vmat_top: material cost of top cap
        """

        if pressure == 0:
            return [self.costrate_steel*x for x in self.get_mass_tower_material(pressure= pressure)] # 2003 dollars
        else:
            Mmat_wall, Mmat_bot, Mmat_top= self.get_mass_tower_material(pressure= pressure)
            # use adjusted pressure cap cost
            return [self.costrate_steel*Mmat_wall, self.costrate_endcap*Mmat_bot, self.costrate_endcap*Mmat_top] # 2003 dollars

    def get_operational_mass_fraction(self):
        """
        get the fraction of stored hydrogen to tower mass
        
        following Kottenstette
        """

        Sut= self.ultimate_tensile_strength
        rho= self.density_steel
        R= self.gasconstant_H2
        T= self.operating_temp + 273.15 # convert to K

        frac= Sut/(rho*R*T)

        return frac # nondim.

    def get_cost_nontower(self,
                          traditional : bool = False,
                          naive : bool = True):
        nonwall_cost= 0
        if traditional:
            nonwall_cost += self.tower_length*self.costrate_ladder # add ladder cost
            nonwall_cost += self.cost_door # add door cost
        else:
            naive= True
            if naive:
                nonwall_cost += self.cost_mainframe_extension
                nonwall_cost += 2*self.cost_door
                nonwall_cost += 2*self.tower_length*self.costrate_ladder
                nonwall_cost += self.cost_nozzles_manway
                nonwall_cost += self.tower_length*self.costrate_conduit
            else:
                # adjust length b.c. conduit, one ladder must ride outside pressure vessel
                adj_length= np.sqrt((self.section_diameters[0] - self.section_diameters[-1])**2 + self.tower_length**2)
                nonwall_cost += self.cost_mainframe_extension
                nonwall_cost += 2*self.cost_door
                nonwall_cost += self.tower_length*self.costrate_ladder
                nonwall_cost += adj_length*self.costrate_ladder
                nonwall_cost += self.cost_nozzles_manway
                nonwall_cost += adj_length*self.costrate_conduit
        return nonwall_cost # 2003 dollars

    ### OFFICIAL OUTPUT INTERFACE

    def get_capex(self):
        """ return the total additional capex necessary for H2 production """
        capex_withH2= self.get_cost_nontower() + np.sum(self.get_cost_tower_material())
        capex_without= self.get_cost_nontower(traditional= True) + np.sum(self.get_cost_tower_material(pressure= 0))
        return capex_withH2 - capex_without # 2003 dollars

    def get_opex(self):
        """
        a simple model for operational expenditures for PV

        maintenance for pressure vessel based on an annual maintenance rate
        against the vessel-specific capital expenditure, plus wages times staff
        hours per year
        """
        
        return self.get_capex()*self.maintenance_rate + self.wage*self.staff_hours # 2003 dollars 

    def get_mass_empty(self):
        """ return the total additional empty mass necessary for H2 production in kg """

        Mtower_withH2= np.sum(self.get_mass_tower_material())
        Mnontower_withH2= 0.0 # not specified

        Mtower_without= np.sum(self.get_mass_tower_material(pressure= 0))
        Mnontower_without= 0.0 # not specified

        Mtotal_withH2= Mtower_withH2 + Mnontower_withH2
        Mtotal_without= Mtower_without + Mnontower_without

        return Mtotal_withH2 - Mtotal_without # kg

    def get_capacity_H2(self):
        """ get the ideal gas H2 capacity in kg """

        Tabs= self.operating_temp + 273.15
        R= self.gasconstant_H2
        p= self.operating_pressure
        V= self.get_volume_tower_inner()

        # ideal gas law
        m_H2= p*V/(R*Tabs)

        return m_H2 # kg

    def get_pressure_H2(self): return self.operating_pressure # Pa, trivial, but for delivery

    ### STATIC FUNCTIONS

    @staticmethod
    def compute_cap_thickness(pressure, diameter, strength_yield,
                              safetyfactor_Sy= 1.5, efficiency_weld= 0.80, constant= 0.10):
        """
        compute the necessary thickness for a pressure vessel cap

        $$
        t= d \\sqrt{\\frac{C P}{S E}}
        $$
        with weld joint efficiency E, allowable stress S, pressure P, diameter
        of pressure action d, edge restraint factor C

        assumed:
            - C= 0.10: Fig-UG-34 of ASME Code S VII, div. 1, via Rao's _Companion
                    Guide to the ASME Boiler and Pressure Vessel Code_ (2009),
                    fig. 21.3. type of sketch (a) assumed
            - E= 0.80: conservatively butt weld, inspected
        
        using the ASME pressure vessel code definitions, and values given in
        Rao _Companion Guide to the ASME Boiler and Pressure Vessel Code_ (2009)
        """

        return diameter*np.sqrt(constant*pressure/(efficiency_weld*strength_yield/safetyfactor_Sy))

    @staticmethod
    def compute_frustum_volume(height, base_diameter, top_diameter):
        """
        return the volume of a frustum (truncated cone)
        """
        return np.pi/12.*height*(base_diameter**2 + base_diameter*top_diameter + top_diameter**2) # volume units

    @staticmethod
    def get_crossover_pressure(welded_joint_efficiency : float,
                               ultimate_tensile_strength : float,
                               d_t_ratio : float):
        """
        get burst/fatigue crossover pressure
        
        following Kottenstette 2003
        """
    
        # convert to nice variables
        E= welded_joint_efficiency # nondim.
        Sut= ultimate_tensile_strength # pressure units
        d_over_t= d_t_ratio # length-per-length; assumed fixed in this study

        p_crossover= 4*E*Sut/(7*d_over_t*(1 - E/7.)) # pressure units

        return p_crossover # pressure units

    @staticmethod
    def get_thickness_increment_const(pressure : float,
                                      ultimate_tensile_strength : float):
        """
        compute Goodman equation-based thickness increment in m

        following Kottenstette 2003
        """

        # convert to text variables
        p= pressure
        # r= diameter/2
        Sut= ultimate_tensile_strength

        alpha_dtp= 0.25*p/Sut

        return alpha_dtp # length per length



'''
    Author: Jamie Kee
    Feb 7, 2023
    Source for most equations is HDSAM3.1, H2 Compressor sheet
    Output is in 2016 USD
'''

from numpy import interp, mean, dot
from math import log10, ceil, log

class Compressor:
    def __init__(self, p_outlet, flow_rate_kg_d, p_inlet=20, n_compressors=2, sizing_safety_factor=1.1):
        '''
            Parameters:
            ---------------
            p_outlet: oulet pressure (bar)
            flow_Rate_kg_d: mass flow rate in kg/day
        '''
        self.p_inlet = p_inlet # bar
        self.p_outlet = p_outlet # bar
        self.flow_rate_kg_d = flow_rate_kg_d # kg/day

        self.n_compressors = n_compressors # At least 2 compressors are recommended for operation at any given time
        self.n_comp_back_up = 1 # Often times, an extra compressor is purchased and installed so that the system can operate at a higher availability.
        self.sizing_safety_factor = sizing_safety_factor # typically oversized. Default to oversize by 10%

        if flow_rate_kg_d*(1/24)*(1/60**2)/n_compressors > 5.4:
            # largest compressors can only do up to about 5.4 kg/s
            """
            H2A Hydrogen Delivery Infrastructure Analysis Models and Conventional Pathway Options Analysis Results
            DE-FG36-05GO15032
            Interim Report
            Nexant, Inc., Air Liquide, Argonne National Laboratory, Chevron Technology Venture, Gas Technology Institute, National Renewable Energy Laboratory, Pacific Northwest National Laboratory, and TIAX LLC
            May 2008
            """
            raise ValueError("Invalid compressor design. Flow rate must be less than 5.4 kg/s per compressor")
            

    def compressor_power(self):
        R = 8.314 #J/mol-K
        T = 25+273.15 #K
        
        cpcv =  1.41 #H2 Cp/Cv ratio
        sizing = self.sizing_safety_factor # 110% based on typical industrial practices
        isentropic_efficiency = 0.88 # 0.88 based on engineering estimation for a reciprocating compressor

        # https://h2tools.org/hyarc/hydrogen-data/hydrogen-compressibility-different-temperatures-and-pressures
        Z_pressures = [1,10,50,100,300,500,1000] # Pressure for z correlation in bar
        Z_z = [1.0006,1.0059,1.0297,1.0601,1.1879,1.3197,1.6454] # H2 Compressibility aty 25C 
        Z = mean(interp([self.p_inlet,self.p_outlet],Z_pressures,Z_z))

        c_ratio_per_stage = 2.1 #based on engineering estimation for a reciprocating compressor operating on hydrogen
        self.stages = ceil((log10(self.p_outlet)-log10(self.p_inlet))/log10(c_ratio_per_stage))
        flow_per_compressor = self.flow_rate_kg_d/self.n_compressors #kg/d
        flow_per_compressor_kg_mols_sec= flow_per_compressor/24/60/60/2.0158 #convert units for power equation
        p_ratio = self.p_outlet/self.p_inlet
        theorhetical_power = Z*flow_per_compressor_kg_mols_sec*R*T*self.stages*(cpcv/(cpcv-1))*((p_ratio)**((cpcv-1)/(self.stages*cpcv))-1) #kW per compressor assumes equal work by all stages and intercooling to oringal temp
        actual_power = theorhetical_power/isentropic_efficiency #kW per compressor
        motor_efficiency = dot([0.00008,-0.0015,0.0061,0.0311,0.7617],[log(actual_power)**x for x in [4,3,2,1,0]])
        self.motor_rating = sizing*actual_power/motor_efficiency #kW per unit
    
    def compressor_system_power(self):
        return self.motor_rating, self.motor_rating*self.n_compressors # [kW] total system power

    def compressor_costs(self):
        n_comp_total = self.n_compressors + self.n_comp_back_up # 2 compressors + 1 backup for reliability
        production_volume_factor = 0.55 # Assume high production volume
        CEPCI = 1.29/1.1 #Convert from 2007 to 2016$

        cost_per_unit = 1962.2*self.motor_rating**0.8225*production_volume_factor*CEPCI
        if self.stages>2:
            cost_per_unit = cost_per_unit * (1+0.2*(self.stages-2))

        install_cost_factor = 2

        direct_capex = cost_per_unit*n_comp_total*install_cost_factor

        land_required = 10000 #m^2 This doesn't change at all in HDSAM...?
        land_cost = 12.35 #$/m2
        land = land_required*land_cost

        other_capital_pct = [0.05,0.1,0.1,0,0.03,0.12] # These are all percentages of direct capex (site,E&D,contingency,licensing,permitting,owners cost)
        other_capital = dot(other_capital_pct,[direct_capex]*len(other_capital_pct)) + land
        
        total_capex = direct_capex + other_capital

        ##
        # O&M
        ##
        CEPCI_labor = 1.15460860725178/1.08789393425957 #2013 to 2016
        labor_cost = 31.6*CEPCI_labor # Bureau of Labor Statistics- NAICS code# 2212 -- 2016$/man-hr
        labor_required = 288*(self.flow_rate_kg_d/100000)**0.25 #hrs/yr #Base case is 3 days per month for a 100,000 MJ/day useable capacity facility. Scaling factor of 0.25 is used for other sized facilities
        labor = labor_required*labor_cost

        electricity_consumption = self.motor_rating*8760*self.n_compressors
        electricity_price = 0.075 #  Maybe need to edit this?
        electricity = electricity_consumption*electricity_price

        #   All are percents of total capital investments, except for maint and repair is % of direct capex, overhead is % of labor
        other_fixed_pct = [0.01,0.01,0.001,0.04,0.5] # These are percentages of total capital investment - [insurance,property taxes,lic perm, maint repairs, overhead]
        other_fixed = dot(other_fixed_pct,[total_capex]*(len(other_fixed_pct)-2)+[direct_capex,labor])

        total_OM = labor+electricity+other_fixed

        return total_capex, total_OM
        
if __name__ == "__main__":
    p_inlet = 20 # bar
    p_outlet = 68 # bar
    flow_rate_kg_d = 9311
    n_compressors = 2

    comp = Compressor(p_outlet,flow_rate_kg_d, p_inlet=p_inlet, n_compressors=n_compressors)
    comp.compressor_power()
    motor_rating, power = comp.compressor_system_power()
    total_capex,total_OM = comp.compressor_costs() #2016$ , 2016$/y
    print("Power (kW): ", power)
    print(f'CAPEX: {round(total_capex,2)} $')
    print(f'Annual operating expense: {round(total_OM,2)} $/yr')

    # CAPEX: 680590.34 $
    # Annual operating expense: 200014.0 $/yr


import pytest
from pathlib import Path
import numpy as np
import copy

import openmdao.api as om

from hopp.simulation import HoppInterface
from hopp.utilities import load_yaml

from greenheart.tools.optimization.openmdao import GreenHeartComponent, HOPPComponent, TurbineDistanceComponent, BoundaryDistanceComponent
from greenheart.tools.optimization.gc_PoseOptimization import PoseOptimization
from greenheart.tools.optimization.gc_run_greenheart import run_greenheart
from greenheart.simulation.greenheart_simulation import GreenHeartSimulationConfig

ROOT_DIR = Path(__file__).parents[2]

solar_resource_file = ROOT_DIR / "resource_files" / "solar" / "35.2018863_-101.945027_psmv3_60_2012.csv"
wind_resource_file = ROOT_DIR / "resource_files" / "wind" / "35.2018863_-101.945027_windtoolkit_2012_60min_80m_100m.srw"
floris_input_filename = Path(__file__).absolute().parent / "inputs" / "floris_input.yaml"
hopp_config_filename = Path(__file__).absolute().parent / "input_files" / "plant" / "hopp_config_wind_wave_solar_battery.yaml"
greenheart_config_filename = Path(__file__).absolute().parent / "input_files" / "plant" / "greenheart_config.yaml"
turbine_config_filename = Path(__file__).absolute().parent / "input_files" / "turbines" / "osw_18MW.yaml"
rtol = 1E-5

def setup_greenheart():
    config = GreenHeartSimulationConfig(
    filename_hopp_config=hopp_config_filename,
    filename_greenheart_config=greenheart_config_filename,
    filename_turbine_config=turbine_config_filename,
    filename_floris_config=floris_input_filename,
    verbose=False,
    show_plots=False,
    save_plots=False,
    output_dir= str(Path(__file__).absolute().parent / "output"),
    use_profast=True,
    post_processing=False,
    incentive_option=1,
    plant_design_scenario=9,
    output_level=7,
    )
    
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_installed_cost_mw"] = 1434000.0 
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_om_per_kw"] = 29.567
    config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"]["om_fixed"][0] = config.hopp_config["config"]["cost_info"]["wind_om_per_kw"]
    # set skip_financial to false for onshore wind
    config.hopp_config["config"]["simulation_options"]["wind"]["skip_financial"] = False

    config.greenheart_config["opt_options"] = {
            "opt_flag": True,
            "general": {
                "folder_output": "output",
                "fname_output": "test_run_greenheart_optimization_mpi",
            },
            "design_variables": {
                "electrolyzer_rating_kw": {
                    "flag": True,
                    "lower": 10000.0,
                    "upper": 200000.0,
                    "units": "kW",
                },
                "pv_capacity_kw": {
                    "flag": True,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW",
                },
                "wave_capacity_kw": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW",
                },
                "battery_capacity_kw": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW",
                },
                "battery_capacity_kwh": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW*h",
                },
                "battery_capacity_kwh": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW*h",
                },
                "turbine_x": {
                    "flag": False,
                    "lower": 0.0,
                    "upper": 1500000.0,
                    "units": "m",
                },
                "turbine_y": {
                    "flag": False,
                    "lower": 0.0,
                    "upper": 1500000.0,
                    "units": "m",
                },
            },
            "constraints": {
                "turbine_spacing": {
                    "flag": False,
                    "lower": 0.0,
                },
                "boundary_distance": {
                    "flag": False,
                    "lower": 0.0,
                },
            "pv_to_platform_area_ratio": {
                "flag": False, 
                "upper": 1.0, # relative size of solar pv area to platform area
                },
            "user": {}
            },
            "merit_figure": "lcoh",
            "merit_figure_user": {
                "name": "lcoh",
                "max_flag": False,
                "ref": 1.0, # value of objective that scales to 1.0
            },
            "driver": {
                "optimization": {
                    "flag": True,
                    "solver": "SLSQP",
                    "tol": 1E-6,
                    "max_major_iter": 1,
                    "max_minor_iter": 2,
                    "gradient_method": "openmdao",
                    # "time_limit": 10, # (sec) optional
                    # "hist_file_name": "snopt_history.txt", # optional
                    "verify_level": -1, # optional
                    "step_calc": None,
                    "form": "forward", # type of finite differences to use, can be one of ["forward", "backward", "central"]
                    "debug_print": False,
                },
                "design_of_experiments": {
                    "flag": False,
                    "run_parallel": False,
                    "generator": "FullFact", # [Uniform, FullFact, PlackettBurman, BoxBehnken, LatinHypercube]
                    "num_samples": 1, # Number of samples to evaluate model at (Uniform and LatinHypercube only)
                    "seed": 2,
                    "levels":  50, #  Number of evenly spaced levels between each design variable lower and upper bound (FullFactorial only)
                    "criterion": None, # [None, center, c, maximin, m, centermaximin, cm, correelation, corr]
                    "iterations": 1,
                    "debug_print": False
                },
                "step_size_study": {
                    "flag": False  
                },
            },
            "recorder": {
                "flag": True,
                "file_name": str(Path(__file__).absolute().parent / "output" / "recorder.sql"),
                "includes": False,
            },
        }

    return config

# @pytest.mark.mpi
def test_run_greenheart_optimize_mpi(subtests):
        
    try:
        from mpi4py import MPI
    except:
        MPI = False

    config = setup_greenheart()

    prob, config = run_greenheart(config, run_only=False)

    if MPI:
        rank = MPI.COMM_WORLD.Get_rank()
    else:
        rank = 0

    if rank == 0:

        cr = om.CaseReader(Path(__file__).absolute().parent / "output" / "recorder.sql")

        # get initial LCOH
        case = cr.get_case(0)
        lcoh_init = case.get_val("lcoh", units='USD/kg')[0]

        # get final LCOH
        case = cr.get_case(-1)
        lcoh_final = case.get_val("lcoh", units='USD/kg')[0]
        
        with subtests.test("lcoh"):
            assert lcoh_final < lcoh_init



## High-Pressure Reverse Osmosis Desalination Model
"""
Python model of High-Pressure Reverse Osmosis Desalination (HPRO).

Reverse Osmosis (RO) is a membrane separation process. No heating or phase change is necessary.
The majority of energy required is for pressurizing the feed water.

A typical RO system is made up of the following basic components:
Pre-treatment: Removes suspended solids and microorganisms through sterilization, fine filtration and adding chemicals to inhibit precipitation.
High-pressure pump: Supplies the pressure needed to enable the water to pass through the membrane (pressure ranges from 54 to 80 bar for seawater).
Membrane Modules: Membrane assembly consists of a pressure vessel and the membrane. Either sprial wound membranes or hollow fiber membranes are used.
Post-treatment: Consists of sterilization, stabilization, mineral enrichment and pH adjustment of product water.
Energy recovery system: A system where a portion of the pressure energy of the brine is recovered.
"""
import sys
import numpy as np
from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals

def RO_desal(net_power_supply_kW, desal_sys_size, useful_life, plant_life, \
    water_recovery_ratio = 0.30, energy_conversion_factor = 4.2, \
    high_pressure_pump_efficency = 0.70, pump_pressure_kPa = 5366,
    energy_recovery = 0.40):

    """
    Calculates the fresh water flow rate (m^3/hr) as
    a function of supplied power (kW) in RO desal.
    Calculats CAPEX (USD), OPEX (USD/yr), annual cash flows
    based on system's rated capacity (m^3/hr).

    param: net_power_supply_kW: (list), hourly power input [kW]

    param: desal_sys_size: Given as desired fresh water flow rate [m^3/hr]

    param: useful_life: useful life of desal system [years]

    param: plant_life: years of plant operation [years]

    Assumed values:
    Common set points from:
    https://www.sciencedirect.com/science/article/abs/pii/S0011916409008443
    water_recovery_ratio = 0.30
    energy_conversion_factor = 4.2
    high_pressure_pump_efficency = 0.70
    pump_pressure_kPa = 5366    (kept static for simplicity. TODO: Modify pressure through RO process)
    energy_recovery = 0.40
    Assumed energy savings by energy recovery device to be 40% of total energy
    https://www.sciencedirect.com/science/article/pii/S0360544210005578?casa_token=aEz_d_LiSgYAAAAA:88Xa6uHMTZee-djvJIF9KkhpuZmwZCLPHNiThmcwv9k9RC3H17JuSoRWI-l92rrTl_E3kO4oOA


    TODO: modify water recovery to vary based on salinity
    SWRO: Sea water Reverse Osmosis, water >18,000 ppm
    SWRO energy_conversion_factor range 2.5 to 4.2 kWh/m^3

    BWRO: Brakish water Reverse Osmosis, water < 18,000 ppm
    BWRO energy_conversion_factor range 1.0 to 1.5 kWh/m^3
    Source: https://www.sciencedirect.com/science/article/pii/S0011916417321057
    """
    # net_power_supply_kW = np.array(net_power_supply_kW)

    desal_power_max = desal_sys_size * energy_conversion_factor #kW

    # Modify power to not exceed system's power maximum (100% rated power capacity) or
    # minimum (approx 50% rated power capacity --> affects filter fouling below this level)
    net_power_for_desal = list()
    operational_flags = list()
    feed_water_flowrate = list()
    fresh_water_flowrate = list()
    for i, power_at_time_step in enumerate(net_power_supply_kW):
        if power_at_time_step > desal_power_max:
            current_net_power_available = desal_power_max
            operational_flag = 2
        elif (0.5 * desal_power_max) <= power_at_time_step <= desal_power_max:
            current_net_power_available = power_at_time_step
            operational_flag = 1
        elif power_at_time_step <= 0.5 * desal_power_max:
            current_net_power_available = 0
            operational_flag = 0

        # Append Operational Flags to a list
        operational_flags.append(operational_flag)
        # Create list of net power available for desal at each timestep
        net_power_for_desal.append(current_net_power_available)

        # Create list of feedwater flowrates based on net power available for desal
        # https://www.sciencedirect.com/science/article/abs/pii/S0011916409008443
        instantaneous_feed_water_flowrate = ((current_net_power_available * (1 + energy_recovery))\
        * high_pressure_pump_efficency) / pump_pressure_kPa * 3600 #m^3/hr

        instantaneous_fresh_water_flowrate = instantaneous_feed_water_flowrate * water_recovery_ratio  # m^3/hr

        feed_water_flowrate.append(instantaneous_feed_water_flowrate)
        fresh_water_flowrate.append(instantaneous_fresh_water_flowrate)

    # print("Fresh water flowrate: ", fresh_water_flowrate, "m^3/hr")
    # print(net_power_for_desal)
    # net_power_supply_kW = np.where(net_power_supply_kW >= desal_power_max, \
    #     desal_power_max, net_power_supply_kW)
    # net_power_supply_kW = np.where(net_power_supply_kW < 0.5 * desal_power_max, \
    #      0, net_power_supply_kW)
    # print("Net power supply after checks: ",net_power_supply_kW, "kW")


    """Values for CAPEX and OPEX given as $/(kg/s)
    Source: https://www.nrel.gov/docs/fy16osti/66073.pdf
    Assumed density of recovered water = 997 kg/m^3"""

    desal_capex = 32894 * (997 * desal_sys_size / 3600) # Output in USD
    # print("Desalination capex: ", desal_capex, " USD")

    desal_opex = 4841 * (997 * desal_sys_size / 3600) # Output in USD/yr
    # print("Desalination opex: ", desal_opex, " USD/yr")

    """
    Assumed useful life = payment period for capital expenditure.
    compressor amortization interest = 3%
    """
    desal_annuals = simple_cash_annuals(plant_life, useful_life,\
            desal_capex,desal_opex, 0.03)
    # a = 0.03
    # desal_annuals = [0] * useful_life

    # desal_amortization = desal_capex * \
    #     ((a*(1+a)**useful_life)/((1+a)**useful_life - 1))

    # for i in range(len(desal_annuals)):
    #     if desal_annuals[i] == 0:
    #         desal_annuals[i] = desal_amortization + desal_opex
    #     return desal_annuals        #[USD]

    return fresh_water_flowrate, feed_water_flowrate, operational_flags, desal_capex, desal_opex, desal_annuals

# Power = np.linspace(0, 100, 100)
# system_size = np.linspace(1,1000,1000)        #m^3/hr

# f = RO_desal(Power,system_size)

# plt.plot(system_size,f,color="C0")
# plt.xlabel("Desalination System Size [m^3/hr]")
# plt.ylabel("Desalination OPEX [USD/yr]")
# plt.show()

if __name__ == '__main__':
    Power = np.array([446,500,183,200,250,100])
    test = RO_desal(Power,300,30,30)
    print(test)


from pytest import approx

from greenheart.simulation.technologies.ammonia import ammonia

grid_prices_dict = {
    "2035": 89.42320514456621,
    "2036": 89.97947569251141,
    "2037": 90.53574624045662,
    "2038": 91.09201678840184,
    "2039": 91.64828733634704,
    "2040": 92.20455788429224,
    "2041": 89.87291235917809,
    "2042": 87.54126683406393,
    "2043": 85.20962130894978,
    "2044": 82.87797578383562,
    "2045": 80.54633025872147,
    "2046": 81.38632144593608,
    "2047": 82.22631263315068,
    "2048": 83.0663038203653,
    "2049": 83.90629500757991,
    "2050": 84.74628619479452,
    "2051": 84.74628619479452,
    "2052": 84.74628619479452,
    "2053": 84.74628619479452,
    "2054": 84.74628619479452,
    "2055": 84.74628619479452,
    "2056": 84.74628619479452,
    "2057": 84.74628619479452,
    "2058": 84.74628619479452,
    "2059": 84.74628619479452,
    "2060": 84.74628619479452,
    "2061": 84.74628619479452,
    "2062": 84.74628619479452,
    "2063": 84.74628619479452,
    "2064": 84.74628619479452,
}

financial_assumptions = {
    "total income tax rate": 0.2574,
    "capital gains tax rate": 0.15,
    "leverage after tax nominal discount rate": 0.10893,
    "debt equity ratio of initial financing": 0.624788,
    "debt interest rate": 0.050049,
}


def test_run_ammonia_model():
    capacity = 100.0
    capacity_factor = 0.9

    ammonia_production_kgpy = ammonia.run_ammonia_model(capacity, capacity_factor)

    assert ammonia_production_kgpy == 90


def test_ammonia_cost_model(subtests):
    config = ammonia.AmmoniaCostModelConfig(
        plant_capacity_kgpy=362560672.27155423,
        plant_capacity_factor=0.9,
        feedstocks=ammonia.Feedstocks(
            electricity_cost=89.42320514456621,
            hydrogen_cost=4.2986685034417045,
            cooling_water_cost=0.00291,
            iron_based_catalyst_cost=23.19977341,
            oxygen_cost=0,
        ),
    )

    res: ammonia.AmmoniaCostModelOutputs = ammonia.run_ammonia_cost_model(config)

    with subtests.test("Total CapEx"):
        assert res.capex_total == approx(74839480.74961768)
    with subtests.test("Fixed OpEx"):
        assert res.total_fixed_operating_cost == approx(10569658.376900101)
    with subtests.test("Variable costs"):
        assert res.variable_cost_in_startup_year == approx(4259805.969069265)
    with subtests.test("Land costs"):
        assert res.land_cost == approx(2160733.0556864925)


def test_ammonia_finance_model():
    cost_config = ammonia.AmmoniaCostModelConfig(
        plant_capacity_kgpy=362560672.27155423,
        plant_capacity_factor=0.9,
        feedstocks=ammonia.Feedstocks(
            electricity_cost=89.42320514456621,
            hydrogen_cost=4.2986685034417045,
            cooling_water_cost=0.00291,
            iron_based_catalyst_cost=23.19977341,
            oxygen_cost=0,
        ),
    )

    costs: ammonia.AmmoniaCostModelOutputs = ammonia.run_ammonia_cost_model(cost_config)

    plant_capacity_kgpy = 362560672.27155423
    plant_capacity_factor = 0.9

    config = ammonia.AmmoniaFinanceModelConfig(
        plant_life=30,
        plant_capacity_kgpy=plant_capacity_kgpy,
        plant_capacity_factor=plant_capacity_factor,
        feedstocks=cost_config.feedstocks,
        grid_prices=grid_prices_dict,
        financial_assumptions=financial_assumptions,
        costs=costs,
    )

    lcoa_expected = 0.9322866176899477

    res: ammonia.AmmoniaFinanceModelOutputs = ammonia.run_ammonia_finance_model(config)

    assert res.sol.get("price") == lcoa_expected


def test_ammonia_size_h2_input(subtests):
    config = ammonia.AmmoniaCapacityModelConfig(
        hydrogen_amount_kgpy=73288888.8888889,
        input_capacity_factor_estimate=0.9,
        feedstocks=ammonia.Feedstocks(
            electricity_cost=89.42320514456621,
            hydrogen_cost=4.2986685034417045,
            cooling_water_cost=0.00291,
            iron_based_catalyst_cost=23.19977341,
            oxygen_cost=0,
        ),
    )

    res: ammonia.AmmoniaCapacityModelOutputs = ammonia.run_size_ammonia_plant_capacity(
        config
    )

    with subtests.test("Ammonia plant size"):
        assert res.ammonia_plant_capacity_kgpy == approx(334339658.8730839)
    with subtests.test("hydrogen input"):
        assert res.hydrogen_amount_kgpy == approx(73288888.8888889)


def test_ammonia_size_NH3_input(subtests):
    config = ammonia.AmmoniaCapacityModelConfig(
        desired_ammonia_kgpy=334339658.8730839,
        input_capacity_factor_estimate=0.9,
        feedstocks=ammonia.Feedstocks(
            electricity_cost=89.42320514456621,
            hydrogen_cost=4.2986685034417045,
            cooling_water_cost=0.00291,
            iron_based_catalyst_cost=23.19977341,
            oxygen_cost=0,
        ),
    )

    res: ammonia.AmmoniaCapacityModelOutputs = ammonia.run_size_ammonia_plant_capacity(
        config
    )

    with subtests.test("Ammonia plant size"):
        assert res.ammonia_plant_capacity_kgpy == approx(371488509.8589821)
    with subtests.test("hydrogen input"):
        assert res.hydrogen_amount_kgpy == approx(73288888.8888889)


def test_ammonia_full_model(subtests):
    config = {
        "ammonia": {
            "capacity": {
                "hydrogen_amount_kgpy": 73288888.8888889,
                "input_capacity_factor_estimate": 0.9,
            },
            "costs": {
                "feedstocks": {
                    "electricity_cost": 89.42320514456621,
                    "hydrogen_cost": 4.2986685034417045,
                    "cooling_water_cost": 0.00291,
                    "iron_based_catalyst_cost": 23.19977341,
                    "oxygen_cost": 0,
                },
            },
            "finances": {
                "plant_life": 30,
                "grid_prices": grid_prices_dict,
                "financial_assumptions": financial_assumptions,
            },
        }
    }

    res = ammonia.run_ammonia_full_model(config)

    assert len(res) == 3

    with subtests.test("Ammonia plant size"):
        assert res[0].ammonia_plant_capacity_kgpy == approx(334339658.8730839)

    with subtests.test("capex"):
        assert res[1].capex_total == approx(71189795.03176591)

    with subtests.test("LCOA"):
        assert res[2].sol.get("price") == approx(0.934951845207612)



from greenheart.to_organize.H2_Analysis.h2_setup_optimize import calculate_h_lcoe
from greenheart.to_organize.H2_Analysis.simple_dispatch import SimpleDispatch
from greenheart.to_organize.gradient_free import GeneticAlgorithm
import pandas as pd
import numpy as np
import warnings
warnings.filterwarnings("ignore")

# h_lcoe, np.sum(combined_hybrid_power_production_hopp), H2_Results['hydrogen_annual_output'], total_system_installed_cost, total_annual_operating_costs

def objective_function(x):
    global bat_model
    global scenario
    global buy_from_grid
    global sell_to_grid
    global best_solution

    electrolyzer_size_mw = x[0]
    solar_capacity_mw = x[1]
    battery_storage_mwh = x[2]
    n_turbines = int(x[3])

    h_lcoe, _, _, _, _ = calculate_h_lcoe(bat_model,electrolyzer_size_mw,n_turbines,solar_capacity_mw,battery_storage_mwh,
                                scenario,buy_from_grid=buy_from_grid,sell_to_grid=sell_to_grid)


    if h_lcoe < best_solution:
        best_solution = h_lcoe
        print("_____________________________")
        print("best_solution: ", h_lcoe)
        print("electrolyzer_size_mw: ", electrolyzer_size_mw)
        print("solar_capacity_mw: ", solar_capacity_mw)
        print("battery_storage_mwh: ", battery_storage_mwh)
        print("n_turbines: ", n_turbines)

    return h_lcoe

def optimize_gf():
    global bat_model
    global scenario
    global buy_from_grid
    global sell_to_grid
    global best_solution

    bat_model = SimpleDispatch()
    scenario = pd.read_csv('single_scenario.csv') 
    buy_from_grid = False
    sell_to_grid = False
    best_solution = 1E16

    ga = GeneticAlgorithm()
    ga.objective_function = objective_function
    ga.bits = np.array([8,8,8,8])
    ga.bounds = np.array([(1E-6,200),(0,200),(0,200),(0,100)])
    ga.variable_type = np.array(["float","float","float","int"])
    
    ga.max_generation = 30
    ga.population_size = 15
    ga.convergence_iters = 10
    ga.tol = 1E-6
    ga.crossover_rate = 0.1
    ga.mutation_rate = 0.01

    ga.optimize_ga(print_progress=False)

    solution_history = ga.solution_history
    opt_lcoh = ga.optimized_function_value
    opt_vars = ga.optimized_design_variables

    opt_electrolyzer_size_mw = opt_vars[0]
    opt_solar_capacity_mw = opt_vars[1]
    opt_battery_storage_mwh = opt_vars[2]
    opt_n_turbines = int(opt_vars[3])
    return opt_electrolyzer_size_mw, opt_solar_capacity_mw, opt_battery_storage_mwh, opt_n_turbines

if __name__=="__main__":
    global bat_model
    global scenario
    global buy_from_grid
    global sell_to_grid
    global best_solution

    bat_model = SimpleDispatch()
    scenario = pd.read_csv('single_scenario.csv')
    buy_from_grid = False
    sell_to_grid = False
    best_solution = 1E16

    ga = GeneticAlgorithm()
    ga.objective_function = objective_function
    ga.bits = np.array([8,8,8,8])
    ga.bounds = np.array([(1E-6,200),(0,200),(0,200),(0,100)])
    ga.variable_type = np.array(["float","float","float","int"])

    ga.max_generation = 30
    ga.population_size = 15
    ga.convergence_iters = 10
    ga.tol = 1E-6
    ga.crossover_rate = 0.1
    ga.mutation_rate = 0.01

    ga.optimize_ga(print_progress=False)

    solution_history = ga.solution_history
    opt_lcoh = ga.optimized_function_value
    opt_vars = ga.optimized_design_variables

    opt_electrolyzer_size_mw = opt_vars[0]
    opt_solar_capacity_mw = opt_vars[1]
    opt_battery_storage_mwh = opt_vars[2]
    opt_n_turbines = int(opt_vars[3])

    print("opt_lcoh: ", opt_lcoh)
    print("opt_electrolyzer: ", opt_electrolyzer_size_mw)
    print("opt_solar: ", opt_solar_capacity_mw)
    print("opt_battery: ", opt_battery_storage_mwh)
    print("opt_n_turbs: ", opt_n_turbines)

    import matplotlib.pyplot as plt
    plt.plot(solution_history)
    plt.show()



from greenheart.simulation.technologies.hydrogen.electrolysis import PEM_tools
from greenheart.simulation.technologies.steel import steel
from greenheart.simulation.technologies.ammonia import ammonia 


def size_electrolyzer_for_end_use(greenheart_config):
    
    hybrid_electricity_estimated_cf = greenheart_config["project_parameters"]["hybrid_electricity_estimated_cf"]
    
    if "ammonia" in list(greenheart_config.keys()):
        feedstocks = ammonia.Feedstocks({'electricity_cost':0,
        'hydrogen_cost':0,'cooling_water_cost':0,
        'iron_based_catalyst_cost':0,'oxygen_cost':0})
        config = ammonia.AmmoniaCapacityModelConfig(
            input_capacity_factor_estimate = greenheart_config["ammonia"]["capacity"]['input_capacity_factor_estimate'],
            feedstocks = feedstocks,
            desired_ammonia_kgpy = greenheart_config["ammonia"]["capacity"]["annual_production_target"],
            )
        output = ammonia.run_size_ammonia_plant_capacity(config)
        
    if "steel" in list(greenheart_config.keys()):
        feedstocks = steel.Feedstocks(natural_gas_prices={})
        config = steel.SteelCapacityModelConfig(
            input_capacity_factor_estimate=greenheart_config["steel"]["capacity"]['input_capacity_factor_estimate'],
            feedstocks = feedstocks,
            desired_steel_mtpy = greenheart_config["steel"]["capacity"]["annual_production_target"],
        )
        output = steel.run_size_steel_plant_capacity(config)
        
    hydrogen_production_capacity_required_kgphr = output.hydrogen_amount_kgpy/(8760*hybrid_electricity_estimated_cf)
    
    deg_power_inc = greenheart_config["electrolyzer"]['eol_eff_percent_loss']/100
    bol_or_eol_sizing = greenheart_config["electrolyzer"]["sizing"]["size_for"]
    cluster_cap_mw = greenheart_config["electrolyzer"]["cluster_rating_MW"]
    electrolyzer_capacity_BOL_MW = PEM_tools.size_electrolyzer_for_hydrogen_demand(hydrogen_production_capacity_required_kgphr, size_for = bol_or_eol_sizing,electrolyzer_degradation_power_increase=deg_power_inc)
    electrolyzer_size_mw = PEM_tools.check_capacity_based_on_clusters(electrolyzer_capacity_BOL_MW,cluster_cap_mw)

    greenheart_config["electrolyzer"]["rating"] = electrolyzer_size_mw
    greenheart_config["electrolyzer"]["sizing"]["hydrogen_dmd"] = hydrogen_production_capacity_required_kgphr

    return greenheart_config

def run_resizing_estimation(greenheart_config):
    if greenheart_config["project_parameters"]["hybrid_electricity_estimated_cf"] > 1:
        raise(ValueError("hybrid plant capacity factor estimate (hybrid_electricity_estimated_cf) cannot exceed 1"))
    
    if greenheart_config["project_parameters"]["grid_connection"]:
        if greenheart_config["project_parameters"]["hybrid_electricity_estimated_cf"]<1:
            print("hybrid_electricity_estimated_cf reset to 1 for grid-connected cases")
            greenheart_config["project_parameters"]["hybrid_electricity_estimated_cf"] = 1
    
    if greenheart_config["electrolyzer"]["sizing"]["resize_for_enduse"]:
        greenheart_config = size_electrolyzer_for_end_use(greenheart_config)

    return greenheart_config




"""
Author: Jamie Kee
Added to HOPP by: Jared Thomas
Note: ANL costs are in 2018 dollars

07/15/2024: Jamie removed Z=0.9 assumption with linear approx, 
removed f=0.01 assumption with Hofer eqn, added
algebraic solver, and reformatted with black.
08/02/2024: Provide cost overrides
"""

import pandas as pd
import numpy as np
import os

BAR2MPA = 0.1
BAR2PA = 100_000
MM2IN = 0.0393701
M2KM = 1 / 1_000
M2MM = 1_000
KM2MI = 0.621371


def run_pipe_analysis(
    L: float,
    m_dot: float,
    p_inlet: float,
    p_outlet: float,
    depth: float,
    risers: int = 1,
    data_location: str = os.path.abspath(os.path.dirname(__file__) + "/data_tables"),
    labor_in_mi: float = None,
    misc_in_mi: float = None,
    row_in_mi: float = None,
    mat_in_mi: float = None,
    region: str = "SW",
):
    """
    This function calculates the cheapest grade, diameter, thickness, subject to ASME B31.12 and .8

    If $/in/mi values are provided in labor_in_mi, misc_in_mi, row_in_mi, mat_in_mi, those values
    will be used in the cost calculations instead of the defaults
    """
    p_inlet_MPa = p_inlet * BAR2MPA
    F = 0.72  # Design option B class 1 - 2011 ASME B31.12 Table  PL-3.7.1.2
    E = 1.0  # Long. Weld Factor: Seamless (Table IX-3B)
    T_derating = 1  # 2020 ASME B31.8 Table A841.1.8-1 for T<250F, 121C

    # Cost overrides
    anl_cost_overrides = {"labor": labor_in_mi, "misc": misc_in_mi, "ROW": row_in_mi}

    riser = (
        risers > 0
    )  # This is a flag for the ASMEB31.8 stress design, if not including risers, then this can be set to false
    extra_length = 1 + 0.05  # 5% extra
    total_L = (
        L * extra_length + risers * depth * M2KM
    )  # km #Assuming 5% extra length and 1 riser. Will need two risers for turbine to central platform

    #   Import mechanical props and pipe thicknesses (remove A,B ,and A25 since no costing data)
    yield_strengths = pd.read_csv(
        os.path.join(data_location, "steel_mechanical_props.csv"),
        index_col=None,
        header=0,
    )
    yield_strengths = yield_strengths.loc[
        ~yield_strengths["Grade"].isin(["A", "B", "A25"])
    ].reset_index()
    schedules_all = pd.read_csv(
        os.path.join(data_location, "pipe_dimensions_metric.csv"),
        index_col=None,
        header=0,
    )
    steel_costs_kg = pd.read_csv(
        os.path.join(data_location, "steel_costs_per_kg.csv"), index_col=None, header=0
    )

    #   First get the minimum diameter required to achieve the outlet pressure for given length and m_dot
    min_diam_mm = get_min_diameter_of_pipe(
        L=L, m_dot=m_dot, p_inlet=p_inlet, p_outlet=p_outlet
    )
    #   Filter for diameters larger than min diam required
    schedules_spec = schedules_all.loc[schedules_all["DN"] >= (min_diam_mm)]

    #   Gather the grades, diameters, and schedules to loop thru
    grades = yield_strengths["Grade"].values
    diams = schedules_spec["Outer diameter [mm]"].values
    schds = schedules_spec.loc[
        :, ~schedules_spec.columns.isin(["DN", "Outer diameter [mm]"])
    ].columns
    viable_types = []

    #   Loop thru grades
    for grade in grades:
        #   Get SMYS and SMTS for the specific grade
        SMYS = yield_strengths.loc[yield_strengths["Grade"] == grade, "SMYS [Mpa]"].iat[
            0
        ]
        SMTS = yield_strengths.loc[yield_strengths["Grade"] == grade, "SMTS [Mpa]"].iat[
            0
        ]
        #   Loop thru outer diameters
        for diam in diams:
            diam_row = schedules_spec.loc[schedules_spec["Outer diameter [mm]"] == diam]
            dn = diam_row["DN"].iat[0]
            #   Loop thru scheudles (which give the thickness)
            for schd in schds:
                thickness = diam_row[schd].iat[0]

                # Check if thickness satisfies ASME B31.12
                mat_perf_factor = get_mat_factor(
                    SMYS=SMYS, SMTS=SMTS, design_pressure=p_inlet * BAR2MPA
                )
                t_ASME = p_inlet_MPa * dn / (2 * SMYS * F * E * mat_perf_factor)
                if thickness < t_ASME:
                    continue

                # Check if satifies ASME B31.8
                if not checkASMEB318(
                    SMYS=SMYS,
                    diam=diam,
                    thickness=thickness,
                    riser=riser,
                    depth=depth,
                    p_inlet=p_inlet,
                    T_derating=T_derating,
                ):
                    continue

                # Add qualified pipes to saved answers:
                inner_diam = diam - 2 * thickness
                viable_types.append([grade, dn, diam, inner_diam, schd, thickness])

    viable_types_df = pd.DataFrame(
        viable_types,
        columns=[
            "Grade",
            "DN",
            "Outer diameter (mm)",
            "Inner diameter (mm)",
            "Schedule",
            "Thickness (mm)",
        ],
    ).dropna()

    #   Calculate material, labor, row, and misc costs
    viable_types_df = get_mat_costs(
        schedules_spec=viable_types_df,
        total_L=total_L,
        steel_costs_kg=steel_costs_kg,
        mat_cost_override=mat_in_mi,
    )
    viable_types_df = get_anl_costs(
        costs=viable_types_df,
        total_L=total_L,
        anl_cost_overrides=anl_cost_overrides,
        loc=region,
    )
    viable_types_df["total capital cost [$]"] = viable_types_df[
        ["mat cost [$]", "labor cost [$]", "misc cost [$]", "ROW cost [$]"]
    ].sum(axis=1)

    # Annual operating cost assumes 1.17% of total capital
    # https://doi.org/10.1016/j.esr.2021.100658
    viable_types_df["annual operating cost [$]"] = (
        0.0117 * viable_types_df["total capital cost [$]"]
    )

    # Take the option with the lowest total capital cost
    min_row = (
        viable_types_df.sort_values(by="total capital cost [$]").iloc[:1].reset_index()
    )
    return min_row


def get_mat_factor(SMYS: float, SMTS: float, design_pressure: float) -> float:
    """
    Determine the material performance factor ASMEB31.12.
    Dependent on the SMYS and SMTS.
    Defaulted to 1 if not within parameters - This may not be a good assumption
    """
    dp_array = np.array(
        [6.8948, 13.7895, 15.685, 16.5474, 17.9264, 19.3053, 20.6843]
    )  # MPa
    if SMYS <= 358.528 or SMTS <= 455.054:
        h_f_array = np.array([1, 1, 0.954, 0.91, 0.88, 0.84, 0.78])
    elif SMYS <= 413.686 and (SMTS > 455.054 and SMTS <= 517.107):
        h_f_array = np.array([0.874, 0.874, 0.834, 0.796, 0.77, 0.734, 0.682])
    elif SMYS <= 482.633 and (SMTS > 517.107 and SMTS <= 565.370):
        h_f_array = np.array([0.776, 0.776, 0.742, 0.706, 0.684, 0.652, 0.606])
    elif SMYS <= 551.581 and (SMTS > 565.370 and SMTS <= 620.528):
        h_f_array = np.array([0.694, 0.694, 0.662, 0.632, 0.61, 0.584, 0.542])
    else:
        return 1
    mat_perf_factor = np.interp(design_pressure, dp_array, h_f_array)
    return mat_perf_factor


def checkASMEB318(
    SMYS: float,
    diam: float,
    thickness: float,
    riser: bool,
    depth: float,
    p_inlet: float,
    T_derating: float,
) -> bool:
    """
    Determine if pipe parameters satisfy hoop and longitudinal stress requirements
    """

    # Hoop Stress - 2020 ASME B31.8 Table A842.2.2-1
    F1 = 0.50 if riser else 0.72

    #   Hoop stress (MPa) - 2020 ASME B31.8 section A842.2.2.2 eqn (1)
    #   This is the maximum value for S_h
    #   Sh <= F1*SMYS*T_derating
    S_h_check = F1 * SMYS * T_derating

    #   Hoop stress (MPa)
    rho_water = 1_000  # kg/m3
    p_hydrostatic = rho_water * 9.81 * depth / BAR2PA  # bar
    dP = (p_inlet - p_hydrostatic) * BAR2MPA  # MPa
    S_h = (
        dP * (diam - (thickness if diam / thickness >= 30 else 0)) / (2_000 * thickness)
    )
    if S_h >= S_h_check:
        return False

    #   Longitudinal stress (MPa)
    S_L_check = 0.8 * SMYS  # 2020 ASME B31.8 Table A842.2.2-1. Same for riser and pipe
    S_L = p_inlet * BAR2MPA * (diam - 2 * thickness) / (4 * thickness)
    if S_L > S_L_check:
        return False

    S_combined_check = (
        0.9 * SMYS
    )  # 2020 ASME B31.8 Table A842.2.2-1. Same for riser and pipe
    #   Torsional stress?? Under what applied torque? Not sure what to do for this.

    return True


def get_anl_costs(
    costs: pd.DataFrame, total_L: float, anl_cost_overrides: dict, loc: str = "SW"
) -> pd.DataFrame:
    """
    Calculates the labor, right-of-way (ROW), and miscellaneous costs associated with pipe capital cost

    Users can specify a region (GP,NE,MA,GL,RM,SE,PN,SW,CA) that corresponds to grouping of states which
    will apply cost correlations from Brown, D., et al. 2022. “The Development of Natural Gas and Hydrogen Pipeline Capital
    Cost Estimating Equations.” International Journal of Hydrogen Energy https://doi.org/10.1016/j.ijhydene.2022.07.270.

    Alternatively, if a value (not None) is provided in anl_cost_overrides, that value be used as the $/in/mi
    cost correlation for the relevant cost type.
    """

    ANL_COEFS = {
        "GP": {
            "labor": [10406, 0.20953, -0.08419],
            "misc": [4944, 0.17351, -0.07621],
            "ROW": [2751, -0.28294, 0.00731],
            "material": [5813, 0.31599, -0.00376],
        },
        "NE": {
            "labor": [249131, -0.33162, -0.17892],
            "misc": [65990, -0.29673, -0.06856],
            "ROW": [83124, -0.66357, -0.07544],
            "material": [10409, 0.296847, -0.07257],
        },
        "MA": {
            "labor": [43692, 0.05683, -0.10108],
            "misc": [14616, 0.16354, -0.16186],
            "ROW": [1942, 0.17394, -0.01555],
            "material": [9113, 0.279875, -0.00840],
        },
        "GL": {
            "labor": [58154, -0.14821, -0.10596],
            "misc": [41238, -0.34751, -0.11104],
            "ROW": [14259, -0.65318, 0.06865],
            "material": [8971, 0.255012, -0.03138],
        },
        "RM": {
            "labor": [10406, 0.20953, -0.08419],
            "misc": [4944, 0.17351, -0.07621],
            "ROW": [2751, -0.28294, 0.00731],
            "material": [5813, 0.31599, -0.00376],
        },
        "SE": {
            "labor": [32094, 0.06110, -0.14828],
            "misc": [11270, 0.19077, -0.13669],
            "ROW": [9531, -0.37284, 0.02616],
            "material": [6207, 0.38224, -0.05211],
        },
        "PN": {
            "labor": [32094, 0.06110, -0.14828],
            "misc": [11270, 0.19077, -0.13669],
            "ROW": [9531, -0.37284, 0.02616],
            "material": [6207, 0.38224, -0.05211],
        },
        "SW": {
            "labor": [95295, -0.53848, 0.03070],
            "misc": [19211, -0.14178, -0.04697],
            "ROW": [72634, -1.07566, 0.05284],
            "material": [5605, 0.41642, -0.06441],
        },
        "CA": {
            "labor": [95295, -0.53848, 0.03070],
            "misc": [19211, -0.14178, -0.04697],
            "ROW": [72634, -1.07566, 0.05284],
            "material": [5605, 0.41642, -0.06441],
        },
    }

    if loc not in ANL_COEFS.keys():
        raise ValueError(f"Region {loc} was supplied, but is not a valid region")

    L_mi = total_L * KM2MI

    def cost_per_in_mi(coef: list, DN_in: float, L_mi: float) -> float:
        return coef[0] * DN_in ** coef[1] * L_mi ** coef[2]

    diam_col = "DN"
    for cost_type in ["labor", "misc", "ROW"]:
        cost_per_in_mi_val = anl_cost_overrides[cost_type]
        # If no override specified, use defaults
        if cost_per_in_mi_val is None:
            cost_per_in_mi_val = costs.apply(
                lambda x: cost_per_in_mi(
                    ANL_COEFS[loc][cost_type], x[diam_col] * MM2IN, L_mi
                ),
                axis=1,
            )
        costs[f"{cost_type} cost [$]"] = (
            cost_per_in_mi_val * costs[diam_col] * MM2IN * L_mi
        )

    return costs


def get_mat_costs(
    schedules_spec: pd.DataFrame,
    total_L: float,
    steel_costs_kg: pd.DataFrame,
    mat_cost_override: float,
):
    """
    Calculates the material cost based on $/kg from Savoy for each grade
    Inc., S. P. Live Stock List & Current Price. https://www.savoypipinginc.com/blog/live-stock-and-current-price.html. Accessed September 22, 2022.

    Users can alternatively provide a $/in/mi override to calculate material cost
    """
    rho_steel = 7840  # kg/m3
    L_m = total_L / M2KM
    L_mi = total_L * KM2MI

    def get_volume(od_mm: float, id_mm: float, L_m: float) -> float:
        return np.pi / 4 * (od_mm**2 - id_mm**2) / M2MM**2 * L_m

    od_col = "Outer diameter (mm)"
    id_col = "Inner diameter (mm)"
    schedules_spec["volume [m3]"] = schedules_spec.apply(
        lambda x: get_volume(x[od_col], x[id_col], L_m),
        axis=1,
    )
    schedules_spec["weight [kg]"] = schedules_spec["volume [m3]"] * rho_steel

    # If mat cost override is not specified, use $/kg savoy costing
    if mat_cost_override is not None:
        schedules_spec["mat cost [$]"] = (
            mat_cost_override * L_mi * schedules_spec["DN"] * MM2IN
        )
    else:
        schedules_spec["mat cost [$]"] = schedules_spec.apply(
            lambda x: x["weight [kg]"]
            * steel_costs_kg.loc[
                steel_costs_kg["Grade"] == x["Grade"], "Price [$/kg]"
            ].iat[0],
            axis=1,
        )

    return schedules_spec


def get_min_diameter_of_pipe(
    L: float, m_dot: float, p_inlet: float, p_outlet: float
) -> float:
    """
    Overview:
    ---------
        This function returns the diameter of a pipe for a given length,flow rate, and pressure boundaries

    Parameters:
    -----------
        L : float - Length of pipeline [km]
        m_dot : float = Mass flow rate [kg/s]
        p_inlet : float = Pressure at inlet of pipe [bar]
        p_outlet : float = Pressure at outlet of pipe [bar]

    Returns:
    --------
        diameter_mm : float - Diameter of pipe [mm]

    """

    p_in_Pa = p_inlet * BAR2PA
    p_out_Pa = p_outlet * BAR2PA

    p_avg = 2 / 3 * (p_in_Pa + p_out_Pa - p_in_Pa * p_out_Pa / (p_in_Pa + p_out_Pa))
    p_diff = (p_in_Pa**2 - p_out_Pa**2) ** 0.5
    T = 15 + 273.15  # Temperature [K]
    R = 8.314  # J/mol-K
    z_fit_params = (6.5466916131e-9, 9.9941320278e-1)  # Slope fit for 15C
    z = z_fit_params[0] * p_avg + z_fit_params[1]
    zrt = z * R * T
    mw = 2.016 / 1_000  # kg/mol for hydrogen
    RO = 0.012  # mm Roughness
    mu = 8.764167e-6  # viscosity
    L_m = L / M2KM

    f_list = [0.01]

    # Diameter depends on Re and f, but are functions of d. So use initial guess
    # of f-0.01, then iteratively solve until f is no longer changing
    err = np.inf
    max_iter = 50
    while err > 0.001:
        d_m = (
            m_dot / p_diff * 4 / np.pi * (mw / zrt / f_list[-1] / L_m) ** (-0.5)
        ) ** (1 / 2.5)
        d_mm = d_m * M2MM
        Re = 4 * m_dot / (np.pi * d_m * mu)
        f_list.append(
            (-2 * np.log10(4.518 / Re * np.log10(Re / 7) + RO / (3.71 * d_mm))) ** (-2)
        )
        err = abs((f_list[-1] - f_list[-2]) / f_list[-2])

        # Error out if no solution after max iterations
        if len(f_list) > max_iter:
            raise ValueError(f"Could not find pipe diameter in {max_iter} iterations")

    return d_mm


if __name__ == "__main__":
    L = 8  # Length [km]
    m_dot = 1.5  # Mass flow rate [kg/s] assuming 300 MW -> 1.5 kg/s
    p_inlet = 30  # Inlet pressure [bar]
    p_outlet = 10  # Outlet pressure [bar]
    depth = 80  # depth of pipe [m]
    costs = run_pipe_analysis(L, m_dot, p_inlet, p_outlet, depth)

    for col in costs.columns:
        print(col, costs[col][0])
